from typing import Any, Dict, List, Tuple

from ._base_response import EC2BaseResponse


def try_parse_int(value: Any, default: Any = None) -> Any:
    try:
        return int(value)
    except (TypeError, ValueError):
        return default


def parse_sg_attributes_from_dict(sg_attributes: Dict[str, Any]) -> Tuple[Any, ...]:
    ip_protocol = sg_attributes.get("IpProtocol", [None])[0]
    from_port = sg_attributes.get("FromPort", [None])[0]
    to_port = sg_attributes.get("ToPort", [None])[0]

    ip_ranges: List[Dict[str, Any]] = []
    ip_ranges_tree = sg_attributes.get("IpRanges") or {}
    for ip_range_idx in sorted(ip_ranges_tree.keys()):
        ip_range = {"CidrIp": ip_ranges_tree[ip_range_idx]["CidrIp"][0]}
        if ip_ranges_tree[ip_range_idx].get("Description"):
            ip_range["Description"] = ip_ranges_tree[ip_range_idx].get("Description")[0]

        ip_ranges.append(ip_range)

    ip_ranges_tree: Dict[str, Any] = sg_attributes.get("Ipv6Ranges") or {}  # type: ignore[no-redef]
    for ip_range_idx in sorted(ip_ranges_tree.keys()):
        ip_range = {"CidrIpv6": ip_ranges_tree[ip_range_idx]["CidrIpv6"][0]}
        if ip_ranges_tree[ip_range_idx].get("Description"):
            ip_range["Description"] = ip_ranges_tree[ip_range_idx].get("Description")[0]

        ip_ranges.append(ip_range)

    if "CidrIp" in sg_attributes:
        cidr_ip = sg_attributes.get("CidrIp")[0]  # type: ignore
        ip_ranges.append({"CidrIp": cidr_ip})

    if "CidrIpv6" in sg_attributes:
        cidr_ipv6 = sg_attributes.get("CidrIpv6")[0]  # type: ignore
        ip_ranges.append({"CidrIpv6": cidr_ipv6})

    source_groups: List[Dict[str, Any]] = []
    groups_tree: Dict[str, Any] = sg_attributes.get("Groups") or {}
    for group_idx in sorted(groups_tree.keys()):
        group_dict = groups_tree[group_idx]
        source_group = {}
        if "GroupId" in group_dict:
            source_group["GroupId"] = group_dict["GroupId"][0]
        if "GroupName" in group_dict:
            source_group["GroupName"] = group_dict["GroupName"][0]
        if "Description" in group_dict:
            source_group["Description"] = group_dict["Description"][0]
        if "OwnerId" in group_dict:
            source_group["OwnerId"] = group_dict["OwnerId"][0]
        source_groups.append(source_group)

    prefix_list_ids: List[Dict[str, Any]] = []
    pl_tree: Dict[str, Any] = sg_attributes.get("PrefixListIds") or {}
    for pl_index in sorted(pl_tree):
        pl_dict = pl_tree.get(pl_index, {})
        pl_item = {}
        if "PrefixListId" in pl_dict:
            pl_item["PrefixListId"] = pl_dict.get("PrefixListId")[0]
        if "Description" in pl_dict:
            pl_item["Description"] = pl_dict.get("Description")[0]
        if pl_item:
            prefix_list_ids.append(pl_item)
    return (ip_protocol, from_port, to_port, ip_ranges, source_groups, prefix_list_ids)


class SecurityGroups(EC2BaseResponse):
    def _process_rules_from_querystring(self) -> Any:
        group_name_or_id = self._get_param("GroupName") or self._get_param("GroupId")
        security_rule_ids = self._get_multi_param("SecurityGroupRuleId")

        querytree: Dict[str, Any] = {}
        for key, value in self.querystring.items():
            key_splitted = key.split(".")
            key_splitted = [try_parse_int(e, e) for e in key_splitted]

            d = querytree
            for subkey in key_splitted[:-1]:
                if subkey not in d:
                    d[subkey] = {}
                d = d[subkey]
            d[key_splitted[-1]] = value

        sg_rule_tags = self._parse_tag_specification().get("security-group-rule", {})

        if "IpPermissions" not in querytree:
            # Handle single rule syntax
            (
                ip_protocol,
                from_port,
                to_port,
                ip_ranges,
                source_groups,
                prefix_list_ids,
            ) = parse_sg_attributes_from_dict(querytree)

            yield {
                "group_name_or_id": group_name_or_id,
                "ip_protocol": ip_protocol,
                "from_port": from_port,
                "to_port": to_port,
                "ip_ranges": ip_ranges,
                "sgrule_tags": sg_rule_tags,
                "source_groups": source_groups,
                "prefix_list_ids": prefix_list_ids,
                "security_rule_ids": security_rule_ids,
            }

        ip_permissions = querytree.get("IpPermissions") or {}
        for ip_permission_idx in sorted(ip_permissions.keys()):
            ip_permission = ip_permissions[ip_permission_idx]

            (
                ip_protocol,
                from_port,
                to_port,
                ip_ranges,
                source_groups,
                prefix_list_ids,
            ) = parse_sg_attributes_from_dict(ip_permission)

            yield {
                "group_name_or_id": group_name_or_id,
                "ip_protocol": ip_protocol,
                "from_port": from_port,
                "to_port": to_port,
                "ip_ranges": ip_ranges,
                "sgrule_tags": sg_rule_tags,
                "source_groups": source_groups,
                "prefix_list_ids": prefix_list_ids,
                "security_rule_ids": security_rule_ids,
            }

    def authorize_security_group_egress(self) -> str:
        self.error_on_dryrun()

        for kwargs in self._process_rules_from_querystring():
            rule, group = self.ec2_backend.authorize_security_group_egress(**kwargs)
        self.ec2_backend.sg_old_egress_ruls[group.id] = group.egress_rules.copy()
        template = self.response_template(AUTHORIZE_SECURITY_GROUP_EGRESS_RESPONSE)
        return template.render(rule=rule, group=group)

    def authorize_security_group_ingress(self) -> str:
        self.error_on_dryrun()

        for kwargs in self._process_rules_from_querystring():
            rule, group = self.ec2_backend.authorize_security_group_ingress(**kwargs)
        self.ec2_backend.sg_old_ingress_ruls[group.id] = group.ingress_rules.copy()
        template = self.response_template(AUTHORIZE_SECURITY_GROUP_INGRESS_RESPONSE)
        return template.render(rule=rule, group=group)

    def create_security_group(self) -> str:
        name = self._get_param("GroupName")
        description = self._get_param("GroupDescription")
        vpc_id = self._get_param("VpcId")
        tags = self._parse_tag_specification().get("security-group", {})

        self.error_on_dryrun()

        group = self.ec2_backend.create_security_group(
            name, description, vpc_id=vpc_id, tags=tags
        )
        if group:
            self.ec2_backend.sg_old_ingress_ruls[group.id] = group.ingress_rules.copy()
            self.ec2_backend.sg_old_egress_ruls[group.id] = group.egress_rules.copy()
        template = self.response_template(CREATE_SECURITY_GROUP_RESPONSE)
        return template.render(group=group)

    def delete_security_group(self) -> str:
        # TODO this should raise an error if there are instances in the group.
        # See
        # http://docs.aws.amazon.com/AWSEC2/latest/APIReference/ApiReference-query-DeleteSecurityGroup.html

        name = self._get_param("GroupName")
        sg_id = self._get_param("GroupId")

        self.error_on_dryrun()

        if name:
            self.ec2_backend.delete_security_group(name)
        elif sg_id:
            self.ec2_backend.delete_security_group(group_id=sg_id)

        return DELETE_GROUP_RESPONSE

    def describe_security_groups(self) -> str:
        groupnames = self._get_multi_param("GroupName")
        group_ids = self._get_multi_param("GroupId")
        filters = self._filters_from_querystring()

        groups = self.ec2_backend.describe_security_groups(
            group_ids=group_ids, groupnames=groupnames, filters=filters
        )

        template = self.response_template(DESCRIBE_SECURITY_GROUPS_RESPONSE)
        return template.render(groups=groups)

    def describe_security_group_rules(self) -> str:
        group_id = self._get_param("GroupId")
        sg_rule_ids = self._get_param("SecurityGroupRuleId.1")
        filters = self._filters_from_querystring()

        self.error_on_dryrun()

        # if sg rule ids are not None then wrap in a list
        # as expected by ec2_backend.describe_security_group_rules
        if sg_rule_ids:
            sg_rule_ids = [sg_rule_ids]

        rules = self.ec2_backend.describe_security_group_rules(
            group_id, sg_rule_ids, filters
        )
        template = self.response_template(DESCRIBE_SECURITY_GROUP_RULES_RESPONSE)
        return template.render(rules=rules)

    def revoke_security_group_egress(self) -> str:
        self.error_on_dryrun()

        for args in self._process_rules_from_querystring():
            # we don't need this parameter to revoke
            del args["sgrule_tags"]
            self.ec2_backend.revoke_security_group_egress(**args)
        return REVOKE_SECURITY_GROUP_EGRESS_RESPONSE

    def revoke_security_group_ingress(self) -> str:
        self.error_on_dryrun()

        for args in self._process_rules_from_querystring():
            # we don't need this parameter to revoke
            del args["sgrule_tags"]
            self.ec2_backend.revoke_security_group_ingress(**args)
        return REVOKE_SECURITY_GROUP_INGRESS_RESPONSE

    def update_security_group_rule_descriptions_ingress(self) -> str:
        for args in self._process_rules_from_querystring():
            # we don't need this parameter to revoke
            del args["sgrule_tags"]
            group = self.ec2_backend.update_security_group_rule_descriptions_ingress(
                **args
            )
        self.ec2_backend.sg_old_ingress_ruls[group.id] = group.ingress_rules.copy()
        return UPDATE_SECURITY_GROUP_RULE_DESCRIPTIONS_INGRESS

    def update_security_group_rule_descriptions_egress(self) -> str:
        for args in self._process_rules_from_querystring():
            # we don't need this parameter to revoke
            del args["sgrule_tags"]
            group = self.ec2_backend.update_security_group_rule_descriptions_egress(
                **args
            )
        self.ec2_backend.sg_old_egress_ruls[group.id] = group.egress_rules.copy()
        return UPDATE_SECURITY_GROUP_RULE_DESCRIPTIONS_EGRESS


CREATE_SECURITY_GROUP_RESPONSE = """<CreateSecurityGroupResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
   <requestId>59dbff89-35bd-4eac-99ed-be587EXAMPLE</requestId>
   <return>true</return>
   <groupId>{{ group.id }}</groupId>
   <tagSet>
    {% for tag in group.get_tags() %}
        <item>
        <key>{{ tag.key }}</key>
        <value>{{ tag.value }}</value>
        </item>
    {% endfor %}
    </tagSet>
</CreateSecurityGroupResponse>"""

DESCRIBE_SECURITY_GROUP_RULES_RESPONSE = """
<DescribeSecurityGroupRulesResponse xmlns="http://ec2.amazonaws.com/doc/2016-11-15/">
  <requestId>{{ request_id }}</requestId>
  <securityGroupRuleSet>
        {% for rule in rules %}
          {% for ip_range in rule.ip_ranges %}
            <item>
                <fromPort>{{ rule.from_port if rule.from_port is not none else -1 }}</fromPort>
                <toPort>{{ rule.to_port if rule.to_port is not none else -1 }}</toPort>
                {% if ip_range['CidrIp'] %}<cidrIpv4>{{ ip_range['CidrIp'] }}</cidrIpv4>{% endif %}
                {% if ip_range['CidrIpv6'] %}<cidrIpv6>{{ ip_range['CidrIpv6'] }}</cidrIpv6>{% endif %}
                <ipProtocol>{{ rule.ip_protocol }}</ipProtocol>
                <groupId>{{ rule.group_id }}</groupId>
                <groupOwnerId>{{ rule.owner_id }}</groupOwnerId>
                {% if ip_range['Description'] %}<description>{{ ip_range['Description'] }}</description>{% endif %}
                <isEgress>{{ 'true' if rule.is_egress else 'false' }}</isEgress>
                <securityGroupRuleId>{{ rule.id }}</securityGroupRuleId>
                {% if (rule.source_groups | length) > 0 %}
                <referencedGroupInfo>
                  <groupId>{{ rule.source_groups[0]['GroupId'] }}</groupId><userId>{{ rule.source_groups[0]['OwnerId'] }}</userId>
                </referencedGroupInfo>
                {% endif %}
                <tagSet>
                {% for tag in rule.get_tags() %}
                    <item>
                      <key>{{ tag.key }}</key>
                      <value>{{ tag.value }}</value>
                    </item>
                {% endfor %}
                </tagSet> 
            </item>
          {% endfor %}
          {% if rule.ip_ranges | length == 0 %}
            <item>
                <fromPort>{{ rule.from_port if rule.from_port is not none else -1 }}</fromPort>
                <toPort>{{ rule.to_port if rule.to_port is not none else -1 }}</toPort>
                <ipProtocol>{{ rule.ip_protocol }}</ipProtocol>
                {% if rule.prefix_list_ids | length > 0 %}<prefixListId>{{ rule.prefix_list_ids[0]['PrefixListId'] }}</prefixListId>{% endif %}
                <groupId>{{ rule.group_id }}</groupId>
                <groupOwnerId>{{ rule.owner_id }}</groupOwnerId>
                <isEgress>{{ 'true' if rule.is_egress else 'false' }}</isEgress>
                <securityGroupRuleId>{{ rule.id }}</securityGroupRuleId>
                {% if (rule.source_groups | length) > 0 %}
                <referencedGroupInfo>
                  <groupId>{{ rule.source_groups[0]['GroupId'] }}</groupId><userId>{{ rule.source_groups[0]['OwnerId'] }}</userId>
                </referencedGroupInfo>
                {% endif %}
                <tagSet>
                {% for tag in rule.get_tags() %}
                    <item>
                      <key>{{ tag.key }}</key>
                      <value>{{ tag.value }}</value>
                    </item>
                {% endfor %}
                </tagSet> 
            </item>
          {% endif %}
        {% endfor %}
  </securityGroupRuleSet>
</DescribeSecurityGroupRulesResponse>"""

DELETE_GROUP_RESPONSE = """<DeleteSecurityGroupResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
  <requestId>59dbff89-35bd-4eac-99ed-be587EXAMPLE</requestId>
  <return>true</return>
</DeleteSecurityGroupResponse>"""

DESCRIBE_SECURITY_GROUPS_RESPONSE = """<DescribeSecurityGroupsResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
   <requestId>59dbff89-35bd-4eac-99ed-be587EXAMPLE</requestId>
   <securityGroupInfo>
      {% for group in groups %}
          <item>
             <ownerId>{{ group.owner_id }}</ownerId>
             <groupId>{{ group.id }}</groupId>
             <groupName>{{ group.name }}</groupName>
             <groupDescription>{{ group.description }}</groupDescription>
             {% if group.vpc_id %}
             <vpcId>{{ group.vpc_id }}</vpcId>
             {% endif %}
             <ipPermissions>
               {% for rule in group.flattened_ingress_rules %}
                    <item>
                       <ipProtocol>{{ rule.ip_protocol }}</ipProtocol>
                       {% if rule.from_port is not none %}
                       <fromPort>{{ rule.from_port }}</fromPort>
                       {% endif %}
                       {% if rule.to_port is not none %}
                       <toPort>{{ rule.to_port }}</toPort>
                       {% endif %}
                       <groups>
                          {% for source_group in rule.source_groups %}
                              <item>
                                 {% if source_group.OwnerId and source_group.OwnerId != "" %}
                                 <userId>{{ source_group.OwnerId }}</userId>
                                 {% endif %}
                                 {% if source_group.GroupId and source_group.GroupId != "" %}
                                 <groupId>{{ source_group.GroupId }}</groupId>
                                 {% endif %}
                                 {% if source_group.GroupName and source_group.GroupName != "" %}
                                 <groupName>{{ source_group.GroupName }}</groupName>
                                 {% endif %}
                                 {% if source_group.Description and source_group.Description != "" %}
                                 <description>{{ source_group.Description }}</description>
                                 {% endif %}
                              </item>
                          {% endfor %}
                       </groups>
                       <ipRanges>
                          {% for ip_range in rule.ip_ranges %}
                             {% if ip_range['CidrIp'] %}
                              <item>
                                    <cidrIp>{{ ip_range['CidrIp'] }}</cidrIp>
                                    {% if ip_range['Description'] %}
                                    <description>{{ ip_range['Description'] }}</description>
                                    {% endif %}
                              </item>
                              {% endif %}
                          {% endfor %}
                       </ipRanges>
                       <ipv6Ranges>
                        {% for ip_range in rule.ip_ranges %}
                            {% if ip_range['CidrIpv6'] %}
                            <item>
                                <cidrIpv6>{{ ip_range['CidrIpv6'] }}</cidrIpv6>
                                {% if ip_range['Description'] %}
                                <description>{{ ip_range['Description'] }}</description>
                                {% endif %}
                            </item>
                            {% endif %}
                        {% endfor %}
                        </ipv6Ranges>
                        <prefixListIds>
                            {% for prefix_list in rule.prefix_list_ids %}
                            <item>
                                <prefixListId>{{ prefix_list.PrefixListId }}</prefixListId>
                                {% if prefix_list.Description %}
                                <description>{{ prefix_list.Description }}</description>
                                {% endif %}
                            </item>
                            {% endfor %}
                       </prefixListIds>
                    </item>
                {% endfor %}
             </ipPermissions>
             <ipPermissionsEgress>
               {% for rule in group.flattened_egress_rules %}
                    <item>
                       <ipProtocol>{{ rule.ip_protocol }}</ipProtocol>
                       {% if rule.from_port is not none %}
                       <fromPort>{{ rule.from_port }}</fromPort>
                       {% endif %}
                       {% if rule.to_port is not none %}
                       <toPort>{{ rule.to_port }}</toPort>
                       {% endif %}
                       <groups>
                          {% for source_group in rule.source_groups %}
                              <item>
                                 {% if source_group.OwnerId and source_group.OwnerId != "" %}
                                 <userId>{{ source_group.OwnerId }}</userId>
                                 {% endif %}
                                 {% if source_group.GroupId and source_group.GroupId != "" %}
                                 <groupId>{{ source_group.GroupId }}</groupId>
                                 {% endif %}
                                 {% if source_group.GroupName and source_group.GroupName != "" %}
                                 <groupName>{{ source_group.GroupName }}</groupName>
                                 {% endif %}
                                 {% if source_group.Description and source_group.Description != "" %}
                                 <description>{{ source_group.Description }}</description>
                                 {% endif %}
                              </item>
                          {% endfor %}
                       </groups>
                       <ipRanges>
                          {% for ip_range in rule.ip_ranges %}
                             {% if ip_range['CidrIp'] %}
                              <item>
                                    <cidrIp>{{ ip_range['CidrIp'] }}</cidrIp>
                                    {% if ip_range['Description'] %}
                                    <description>{{ ip_range['Description'] }}</description>
                                    {% endif %}
                              </item>
                              {% endif %}
                          {% endfor %}
                       </ipRanges>
                       <ipv6Ranges>
                        {% for ip_range in rule.ip_ranges %}
                            {% if ip_range['CidrIpv6'] %}
                            <item>
                                    <cidrIpv6>{{ ip_range['CidrIpv6'] }}</cidrIpv6>
                                    {% if ip_range['Description'] %}
                                    <description>{{ ip_range['Description'] }}</description>
                                    {% endif %}
                            </item>
                            {% endif %}
                        {% endfor %}
                        </ipv6Ranges>
                        <prefixListIds>
                            {% for prefix_list in rule.prefix_list_ids %}
                            <item>
                                <prefixListId>{{ prefix_list.PrefixListId }}</prefixListId>
                                {% if prefix_list.Description %}
                                <description>{{ prefix_list.Description }}</description>
                                {% endif %}
                            </item>
                            {% endfor %}
                        </prefixListIds>
                    </item>
               {% endfor %}
             </ipPermissionsEgress>
             <tagSet>
               {% for tag in group.get_tags() %}
                 <item>
                   <key>{{ tag.key }}</key>
                   <value>{{ tag.value }}</value>
                 </item>
               {% endfor %}
             </tagSet>
          </item>
      {% endfor %}
   </securityGroupInfo>
</DescribeSecurityGroupsResponse>"""

AUTHORIZE_SECURITY_GROUP_INGRESS_RESPONSE = """<AuthorizeSecurityGroupIngressResponse xmlns="http://ec2.amazonaws.com/doc/2016-11-15/">
    <requestId>b1f67202-c2c2-4ba4-8464-c8b1d8f5af7a</requestId>
    <return>true</return>
    <securityGroupRuleSet>
    {% for item in rule.ip_ranges %}
        <item>
            {% if item.CidrIp %}
            <cidrIpv4>{{ item.CidrIp }}</cidrIpv4>
            {% endif %}
            {% if item.CidrIpv6 %}
            <cidrIpv6>{{ item.CidrIpv6 }}</cidrIpv6>
            {% endif %}
            <description>{{ item.Description or '' }}</description>
            {% if rule.from_port is not none %}
            <fromPort>{{ rule.from_port }}</fromPort>
            {% endif %}
            <groupId>{{ group.id }}</groupId>
            <groupOwnerId>{{ rule.owner_id }}</groupOwnerId>
            <ipProtocol>{{ rule.ip_protocol }}</ipProtocol>
            <isEgress>false</isEgress>
            <securityGroupRuleId>{{ rule.id }}</securityGroupRuleId>
            {% if rule.to_port is not none %}
            <toPort>{{ rule.to_port }}</toPort>
            {% endif %}
            <tagSet>
                {% for tag in rule.get_tags() %}
                    <item>
                      <key>{{ tag.key }}</key>
                      <value>{{ tag.value }}</value>
                    </item>
                {% endfor %}
            </tagSet> 
        </item>
    {% endfor %}
    {% for item in rule.prefix_list_ids %}
        <item>
            <prefixListId>{{ item.PrefixListId }}</prefixListId>
            <description>{{ item.Description or '' }}</description>
            {% if rule.from_port is not none %}
            <fromPort>{{ rule.from_port }}</fromPort>
            {% endif %}
            <groupId>{{ group.id }}</groupId>
            <groupOwnerId>{{ rule.owner_id }}</groupOwnerId>
            <ipProtocol>{{ rule.ip_protocol }}</ipProtocol>
            <isEgress>false</isEgress>
            <securityGroupRuleId>{{ rule.id }}</securityGroupRuleId>
            {% if rule.to_port is not none %}
            <toPort>{{ rule.to_port }}</toPort>
            {% endif %}
        </item>
    {% endfor %}
    {% for item in rule.source_groups %}
        <item>
            {% if item.Description and item.Description != "" %}
            <description>{{ item.Description }}</description>
            {% endif %}
            {% if rule.from_port is not none %}
            <fromPort>{{ rule.from_port }}</fromPort>
            {% endif %}
            <groupId>{{ group.id }}</groupId>
            <groupOwnerId>{{ rule.owner_id }}</groupOwnerId>
            <ipProtocol>{{ rule.ip_protocol }}</ipProtocol>
            <isEgress>true</isEgress>
            <securityGroupRuleId>{{ rule.id }}</securityGroupRuleId>
            {% if rule.to_port is not none %}
            <toPort>{{ rule.to_port }}</toPort>
            {% endif %}
            <referencedGroupInfo>
                {% if item.OwnerId and item.OwnerId != "" %}
                <userId>{{ item.OwnerId }}</userId>
                {% endif %}
                {% if item.GroupId and item.GroupId != "" %}
                <groupId>{{ item.GroupId }}</groupId>
                {% endif %}
                {% if item.VpcId and item.VpcId != "" %}
                <vpcId>{{ item.VpcId }}</vpcId>
                {% endif %}
            </referencedGroupInfo>
        </item>
    {% endfor %}
    </securityGroupRuleSet>
</AuthorizeSecurityGroupIngressResponse>"""

REVOKE_SECURITY_GROUP_INGRESS_RESPONSE = """<RevokeSecurityGroupIngressResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
  <requestId>59dbff89-35bd-4eac-99ed-be587EXAMPLE</requestId>
  <return>true</return>
</RevokeSecurityGroupIngressResponse>"""

AUTHORIZE_SECURITY_GROUP_EGRESS_RESPONSE = """<AuthorizeSecurityGroupEgressResponse xmlns="http://ec2.amazonaws.com/doc/2016-11-15/">
    <requestId>b1f67202-c2c2-4ba4-8464-c8b1d8f5af7a</requestId>
    <return>true</return>
    <securityGroupRuleSet>
    {% for item in rule.ip_ranges %}
        <item>
            {% if item.CidrIp %}
            <cidrIpv4>{{ item.CidrIp }}</cidrIpv4>
            {% endif %}
            {% if item.CidrIpv6 %}
            <cidrIpv6>{{ item.CidrIpv6 }}</cidrIpv6>
            {% endif %}
            <description>{{ item.Description or '' }}</description>
            {% if rule.from_port is not none %}
            <fromPort>{{ rule.from_port }}</fromPort>
            {% endif %}
            <groupId>{{ group.id }}</groupId>
            <groupOwnerId>{{ rule.owner_id }}</groupOwnerId>
            <ipProtocol>{{ rule.ip_protocol }}</ipProtocol>
            <isEgress>true</isEgress>
            <securityGroupRuleId>{{ rule.id }}</securityGroupRuleId>
            {% if rule.to_port is not none %}
            <toPort>{{ rule.to_port }}</toPort>
            {% endif %}
            <tagSet>
                {% for tag in rule.get_tags() %}
                    <item>
                      <key>{{ tag.key }}</key>
                      <value>{{ tag.value }}</value>
                    </item>
                {% endfor %}
            </tagSet> 
        </item>
    {% endfor %}
    {% for item in rule.prefix_list_ids %}
        <item>
            <prefixListId>{{ item.PrefixListId }}</prefixListId>
            <description>{{ item.Description or '' }}</description>
            {% if rule.from_port is not none %}
            <fromPort>{{ rule.from_port }}</fromPort>
            {% endif %}
            <groupId>{{ group.id }}</groupId>
            <groupOwnerId>{{ rule.owner_id }}</groupOwnerId>
            <ipProtocol>{{ rule.ip_protocol }}</ipProtocol>
            <isEgress>true</isEgress>
            <securityGroupRuleId>{{ rule.id }}</securityGroupRuleId>
            {% if rule.to_port is not none %}
            <toPort>{{ rule.to_port }}</toPort>
            {% endif %}
        </item>
    {% endfor %}
    {% for item in rule.source_groups %}
        <item>
            {% if item.Description and item.Description != "" %}
            <description>{{ item.Description }}</description>
            {% endif %}
            {% if rule.from_port is not none %}
            <fromPort>{{ rule.from_port }}</fromPort>
            {% endif %}
            <groupId>{{ group.id }}</groupId>
            <groupOwnerId>{{ rule.owner_id }}</groupOwnerId>
            <ipProtocol>{{ rule.ip_protocol }}</ipProtocol>
            <isEgress>true</isEgress>
            <securityGroupRuleId>{{ rule.id }}</securityGroupRuleId>
            {% if rule.to_port is not none %}
            <toPort>{{ rule.to_port }}</toPort>
            {% endif %}
            <referencedGroupInfo>
                {% if item.OwnerId and item.OwnerId != "" %}
                <userId>{{ item.OwnerId }}</userId>
                {% endif %}
                {% if item.GroupId and item.GroupId != "" %}
                <groupId>{{ item.GroupId }}</groupId>
                {% endif %}
                {% if item.VpcId and item.VpcId != "" %}
                <vpcId>{{ item.VpcId }}</vpcId>
                {% endif %}
            </referencedGroupInfo>
        </item>
    {% endfor %}
    </securityGroupRuleSet>
</AuthorizeSecurityGroupEgressResponse>"""

REVOKE_SECURITY_GROUP_EGRESS_RESPONSE = """<RevokeSecurityGroupEgressResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
  <requestId>59dbff89-35bd-4eac-99ed-be587EXAMPLE</requestId>
  <return>true</return>
</RevokeSecurityGroupEgressResponse>"""

UPDATE_SECURITY_GROUP_RULE_DESCRIPTIONS_INGRESS = """<UpdateSecurityGroupRuleDescriptionsIngressResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
  <requestId>59dbff89-35bd-4eac-99ed-be587EXAMPLE</requestId>
  <return>true</return>
</UpdateSecurityGroupRuleDescriptionsIngressResponse>"""

UPDATE_SECURITY_GROUP_RULE_DESCRIPTIONS_EGRESS = """<UpdateSecurityGroupRuleDescriptionsEgressResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
  <requestId>59dbff89-35bd-4eac-99ed-be587EXAMPLE</requestId>
  <return>true</return>
</UpdateSecurityGroupRuleDescriptionsEgressResponse>"""



grpc_route_spec = {
    "priority": 1,
    "grpcRoute": {
        "action": {
            "weightedTargets": [
                {"port": 8080, "virtualNode": "my-virtual-node", "weight": 50}
            ]
        },
        "match": {
            "metadata": [
                {
                    "invert": False,
                    "match": {"exact": "example-value"},
                    "name": "my-metadata-key",
                }
            ],
            "methodName": "myMethod",
            "port": 8080,
            "serviceName": "myService",
        },
        "retryPolicy": {
            "grpcRetryEvents": ["unavailable", "resource-exhausted"],
            "httpRetryEvents": ["gateway-error"],
            "maxRetries": 3,
            "perRetryTimeout": {"unit": "ms", "value": 200},
            "tcpRetryEvents": ["connection-error"],
        },
        "timeout": {
            "idle": {"unit": "s", "value": 60},
            "perRequest": {"unit": "s", "value": 5},
        },
    },
}

http_route_spec = {
    "priority": 2,
    "httpRoute": {
        "action": {
            "weightedTargets": [
                {"port": 80, "virtualNode": "web-server-node", "weight": 100}
            ]
        },
        "match": {
            "headers": [
                {
                    "invert": True,
                    "match": {"prefix": "Bearer "},
                    "name": "Authorization",
                }
            ],
            "method": "POST",
            "path": {"exact": "/login"},
            "port": 80,
            "queryParameters": [
                {"match": {"exact": "example-match"}, "name": "http-query-param"}
            ],
            "scheme": "http",
        },
        "retryPolicy": {
            "httpRetryEvents": ["gateway-error", "client-error"],
            "maxRetries": 0,
            "perRetryTimeout": {"unit": "ms", "value": 0},
            "tcpRetryEvents": ["connection-error"],
        },
        "timeout": {
            "idle": {"unit": "s", "value": 15},
            "perRequest": {"unit": "s", "value": 1},
        },
    },
}

http2_route_spec = {
    "priority": 3,
    "http2Route": {
        "action": {
            "weightedTargets": [
                {"port": 80, "virtualNode": "web-server-node", "weight": 75}
            ]
        },
        "match": {
            "headers": [
                {
                    "invert": False,
                    "match": {"exact": "application/json"},
                    "name": "Content-Type",
                }
            ],
            "method": "GET",
            "path": {"exact": "/api/products"},
            "port": 80,
            "prefix": "/api",
            "queryParameters": [
                {"match": {"exact": "electronics"}, "name": "category"}
            ],
            "scheme": "https",
        },
        "retryPolicy": {
            "httpRetryEvents": ["server-error"],
            "maxRetries": 2,
            "perRetryTimeout": {"unit": "ms", "value": 500},
            "tcpRetryEvents": ["connection-error"],
        },
        "timeout": {
            "idle": {"unit": "s", "value": 30},
            "perRequest": {"unit": "s", "value": 2},
        },
    },
}

tcp_route_spec = {
    "priority": 4,
    "tcpRoute": {
        "action": {
            "weightedTargets": [
                {"port": 22, "virtualNode": "ssh-server-node", "weight": 100}
            ]
        },
        "match": {"port": 22},
        "timeout": {"idle": {"unit": "s", "value": 600}},
    },
}

modified_http_route_spec = {
    "priority": 5,
    "httpRoute": {
        "action": {
            "weightedTargets": [
                {"port": 8080, "virtualNode": "api-server-node", "weight": 50}
            ]
        },
        "match": {
            "headers": [
                {
                    "invert": False,
                    "match": {"prefix": "Token "},
                    "name": "X-Auth-Token",
                }
            ],
            "method": "GET",
            "path": {"exact": "/profile"},
            "port": 443,
            "queryParameters": [
                {"match": {"exact": "modified-match"}, "name": "filter-param"}
            ],
            "scheme": "https",
        },
        "retryPolicy": {
            "httpRetryEvents": ["server-error"],
            "maxRetries": 3,
            "perRetryTimeout": {"unit": "s", "value": 2},
            "tcpRetryEvents": ["connection-reset"],
        },
        "timeout": {
            "idle": {"unit": "m", "value": 5},
            "perRequest": {"unit": "ms", "value": 500},
        },
    },
}

http_virtual_node_spec = {
    "backendDefaults": {
        "clientPolicy": {
            "tls": {
                "certificate": {
                    "file": {
                        "certificateChain": "/path/to/cert_chain.pem",
                        "privateKey": "/path/to/private_key.pem",
                    }
                },
                "enforce": True,
                "ports": [443],
                "validation": {
                    "subjectAlternativeNames": {
                        "match": {"exact": ["www.example.com", "api.example.com"]}
                    },
                    "trust": {"file": {"certificateChain": "/path/to/ca_bundle.pem"}},
                },
            }
        }
    },
    "backends": [
        {
            "virtualService": {
                "clientPolicy": {
                    "tls": {
                        "enforce": False,
                        "validation": {
                            "subjectAlternativeNames": {
                                "match": {
                                    "exact": [
                                        "example-alternative-name",
                                    ]
                                }
                            },
                            "trust": {
                                "file": {
                                    "certificateChain": "example-certificate-chain"
                                },
                            },
                        },
                    }
                },
                "virtualServiceName": "my-service.default.svc.cluster.local",
            }
        }
    ],
    "listeners": [
        {
            "connectionPool": {
                "http": {"maxConnections": 1000, "maxPendingRequests": 5000}
            },
            "healthCheck": {
                "healthyThreshold": 2,
                "intervalMillis": 5000,
                "path": "/health",
                "port": 80,
                "protocol": "http",
                "timeoutMillis": 2000,
                "unhealthyThreshold": 3,
            },
            "outlierDetection": {
                "baseEjectionDuration": {"unit": "s", "value": 30},
                "interval": {"unit": "s", "value": 10},
                "maxEjectionPercent": 10,
                "maxServerErrors": 5,
            },
            "portMapping": {"port": 80, "protocol": "http"},
            "timeout": {
                "http": {
                    "idle": {"unit": "s", "value": 60},
                    "perRequest": {"unit": "s", "value": 5},
                }
            },
            "tls": {
                "certificate": {
                    "acm": {
                        "certificateArn": "arn:aws:acm:us-east-1:123456789012:certificate/abcdefg-1234-5678-90ab-cdef01234567"
                    }
                },
                "mode": "STRICT",
                "validation": {"trust": {"sds": {"secretName": "my-ca-bundle-secret"}}},
            },
        }
    ],
    "logging": {
        "accessLog": {
            "file": {
                "format": {
                    "json": [
                        {"key": "start_time", "value": "%START_TIME%"},
                        {"key": "method", "value": "%REQ(:METHOD)%"},
                    ]
                },
                "path": "/var/log/appmesh/access.log",
            }
        }
    },
    "serviceDiscovery": {
        "awsCloudMap": {
            "attributes": [{"key": "env", "value": "prod"}],
            "ipPreference": "IPv4_PREFERRED",
            "namespaceName": "my-namespace",
            "serviceName": "my-service",
        }
    },
}

http2_virtual_node_spec = {
    "backendDefaults": {
        "clientPolicy": {
            "tls": {
                "certificate": {
                    "file": {
                        "certificateChain": "/path/to/cert_chain.pem",
                        "privateKey": "/path/to/private_key.pem",
                    }
                },
                "enforce": True,
                "ports": [443],
                "validation": {
                    "subjectAlternativeNames": {
                        "match": {"exact": ["http2.example.com"]}
                    },
                    "trust": {"file": {"certificateChain": "/path/to/ca_bundle.pem"}},
                },
            }
        }
    },
    "backends": [
        {
            "virtualService": {
                "clientPolicy": {
                    "tls": {
                        "enforce": True,
                        "ports": [443],
                        "validation": {
                            "subjectAlternativeNames": {
                                "match": {
                                    "exact": [
                                        "match-me",
                                    ]
                                }
                            },
                            "trust": {"sds": {"secretName": "example-secret-name"}},
                        },
                    }
                },
                "virtualServiceName": "my-http2-service.default.svc.cluster.local",
            }
        }
    ],
    "listeners": [
        {
            "connectionPool": {"http2": {"maxRequests": 1000}},
            "healthCheck": {
                "healthyThreshold": 2,
                "intervalMillis": 5000,
                "path": "/",
                "port": 443,
                "protocol": "http2",
                "timeoutMillis": 2000,
                "unhealthyThreshold": 3,
            },
            "portMapping": {
                "port": 443,
                "protocol": "http2",
            },
            "timeout": {
                "http2": {
                    "idle": {
                        "unit": "s",
                        "value": 120,
                    },
                    "perRequest": {"unit": "s", "value": 10},
                }
            },
            "tls": {
                "certificate": {
                    "acm": {
                        "certificateArn": "arn:aws:acm:us-east-1:123456789012:certificate/abcdefg-1234-5678-90ab-cdef01234567"
                    }
                },
                "mode": "STRICT",
                "validation": {"trust": {"sds": {"secretName": "my-ca-bundle-secret"}}},
            },
        }
    ],
    "logging": {
        "accessLog": {
            "file": {
                "format": {
                    "json": [
                        {"key": "end_time", "value": "%END_TIME%"},
                        {"key": "status_code", "value": "%RESPONSE_CODE%"},
                    ]
                },
                "path": "/var/log/appmesh/new_access.log",
            }
        }
    },
    "serviceDiscovery": {
        "awsCloudMap": {
            "attributes": [{"key": "region", "value": "us-east-1"}],
            "ipPreference": "IPv6_PREFERRED",
            "namespaceName": "new-namespace",
            "serviceName": "new-service",
        }
    },
}

grpc_virtual_node_spec = {
    "backendDefaults": {
        "clientPolicy": {
            "tls": {
                "certificate": {
                    "file": {
                        "certificateChain": "/path/to/cert_chain.pem",
                        "privateKey": "/path/to/private_key.pem",
                    }
                },
                "enforce": True,
                "ports": [443],
                "validation": {
                    "subjectAlternativeNames": {
                        "match": {"exact": ["grpc.example.com"]}
                    },
                    "trust": {"file": {"certificateChain": "/path/to/ca_bundle.pem"}},
                },
            }
        }
    },
    "backends": [
        {
            "virtualService": {
                "clientPolicy": {
                    "tls": {
                        "enforce": True,
                        "ports": [443],
                        "certificate": {
                            "file": {
                                "certificateChain": "grpc-backend-certificate-chain",
                                "privateKey": "grpc-backend-certificate-private-key",
                            },
                        },
                        "validation": {
                            "subjectAlternativeNames": {
                                "match": {
                                    "exact": [
                                        "validation-alternate-name",
                                    ]
                                }
                            },
                            "trust": {
                                "acm": {
                                    "certificateAuthorityArns": [
                                        "example-acm-arn",
                                    ]
                                },
                            },
                        },
                    }
                },
                "virtualServiceName": "my-grpc-service.default.svc.cluster.local",
            }
        }
    ],
    "listeners": [
        {
            "connectionPool": {"grpc": {"maxRequests": 500}},
            "healthCheck": {
                "healthyThreshold": 2,
                "intervalMillis": 5000,
                "port": 50051,
                "protocol": "grpc",
                "timeoutMillis": 2000,
                "unhealthyThreshold": 3,
            },
            "portMapping": {
                "port": 50051,
                "protocol": "grpc",
            },
            "timeout": {
                "grpc": {
                    "idle": {
                        "unit": "s",
                        "value": 600,
                    },
                    "perRequest": {"unit": "s", "value": 30},
                }
            },
            "tls": {
                "certificate": {
                    "acm": {
                        "certificateArn": "arn:aws:acm:us-east-1:123456789012:certificate/abcdefg-1234-5678-90ab-cdef01234567"
                    }
                },
                "mode": "STRICT",
                "validation": {"trust": {"sds": {"secretName": "my-ca-bundle-secret"}}},
            },
        }
    ],
    "logging": {
        "accessLog": {
            "file": {
                "format": {
                    "json": [
                        {"key": "end_time", "value": "%END_TIME%"},
                        {"key": "status_code", "value": "%RESPONSE_CODE%"},
                    ]
                },
                "path": "/var/log/appmesh/new_access.log",
            }
        }
    },
    "serviceDiscovery": {
        "awsCloudMap": {
            "attributes": [{"key": "region", "value": "us-east-1"}],
            "ipPreference": "IPv6_PREFERRED",
            "namespaceName": "new-namespace",
            "serviceName": "new-service",
        }
    },
}

tcp_virtual_node_spec = {
    "backendDefaults": {
        "clientPolicy": {
            "tls": {
                "certificate": {
                    "file": {
                        "certificateChain": "/path/to/cert_chain.pem",
                        "privateKey": "/path/to/private_key.pem",
                    }
                },
                "enforce": True,
                "ports": [443],
                "validation": {
                    "subjectAlternativeNames": {
                        "match": {"exact": ["tcp.example.com"]}
                    },
                    "trust": {"file": {"certificateChain": "/path/to/ca_bundle.pem"}},
                },
            }
        }
    },
    "backends": [
        {
            "virtualService": {
                "clientPolicy": {
                    "tls": {
                        "enforce": False,
                        "validation": {
                            "subjectAlternativeNames": {
                                "match": {
                                    "exact": [
                                        "exact-match-example",
                                    ]
                                }
                            },
                            "trust": {
                                "file": {"certificateChain": "test-certificate-chain"},
                            },
                        },
                    }
                },
                "virtualServiceName": "my-tcp-service.default.svc.cluster.local",
            }
        }
    ],
    "listeners": [
        {
            "connectionPool": {"tcp": {"maxConnections": 2000}},
            "healthCheck": {
                "healthyThreshold": 2,
                "intervalMillis": 10000,
                "port": 8080,
                "protocol": "tcp",
                "timeoutMillis": 5000,
                "unhealthyThreshold": 3,
            },
            "portMapping": {
                "port": 8080,
                "protocol": "tcp",
            },
            "timeout": {
                "tcp": {
                    "idle": {
                        "unit": "m",
                        "value": 30,
                    }
                }
            },
        }
    ],
    "logging": {
        "accessLog": {
            "file": {
                "format": {
                    "json": [
                        {"key": "end_time", "value": "%END_TIME%"},
                        {"key": "status_code", "value": "%RESPONSE_CODE%"},
                    ]
                },
                "path": "/var/log/appmesh/new_access.log",
            }
        }
    },
    "serviceDiscovery": {
        "awsCloudMap": {
            "attributes": [{"key": "region", "value": "us-east-1"}],
            "ipPreference": "IPv6_PREFERRED",
            "namespaceName": "new-namespace",
            "serviceName": "new-service",
        }
    },
}

modified_http2_virtual_node_spec = {
    "backendDefaults": {
        "clientPolicy": {
            "tls": {
                "certificate": {
                    "file": {
                        "certificateChain": "/updated/path/to/cert_chain.pem",
                        "privateKey": "/updated/path/to/private_key.pem",
                    }
                },
                "enforce": False,
                "ports": [8443],
                "validation": {
                    "subjectAlternativeNames": {
                        "match": {"exact": ["updated.example.com"]}
                    },
                    "trust": {
                        "file": {"certificateChain": "/updated/path/to/ca_bundle.pem"}
                    },
                },
            }
        }
    },
    "backends": [
        {
            "virtualService": {
                "clientPolicy": {
                    "tls": {
                        "enforce": False,
                        "ports": [8443],
                        "validation": {
                            "subjectAlternativeNames": {
                                "match": {
                                    "exact": [
                                        "another-exact-match-example",
                                    ]
                                }
                            },
                            "trust": {
                                "file": {
                                    "certificateChain": "different-test-certificate-chain"
                                },
                            },
                        },
                    }
                },
                "virtualServiceName": "updated-http2-service.default.svc.cluster.local",
            }
        }
    ],
    "listeners": [
        {
            "connectionPool": {"http2": {"maxRequests": 500}},
            "healthCheck": {
                "healthyThreshold": 3,
                "intervalMillis": 5100,
                "path": "/health",
                "port": 8443,
                "protocol": "http2",
                "timeoutMillis": 2200,
                "unhealthyThreshold": 2,
            },
            "portMapping": {
                "port": 8443,
                "protocol": "http2",
            },
            "timeout": {
                "http2": {
                    "idle": {
                        "unit": "m",
                        "value": 5,
                    },
                    "perRequest": {"unit": "m", "value": 1},
                }
            },
            "tls": {
                "certificate": {
                    "acm": {
                        "certificateArn": "arn:aws:acm:us-west-2:987654321098:certificate/hgfedcba-4321-8765-ba09-fedc09876543"
                    }
                },
                "mode": "PERMISSIVE",
                "validation": {
                    "trust": {"sds": {"secretName": "updated-ca-bundle-secret"}}
                },
            },
        }
    ],
    "logging": {
        "accessLog": {
            "file": {
                "format": {
                    "json": [
                        {"key": "start_time", "value": "%START_TIME%"},
                        {"key": "method", "value": "%REQUEST_METHOD%"},
                    ]
                },
                "path": "/var/log/appmesh/updated_access.log",
            }
        }
    },
    "serviceDiscovery": {
        "awsCloudMap": {
            "attributes": [{"key": "environment", "value": "production"}],
            "ipPreference": "IPv4_PREFERRED",
            "namespaceName": "updated-namespace",
            "serviceName": "updated-service",
        }
    },
}



import json

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws
from tests import EXAMPLE_AMI_ID

template = """{
  "AWSTemplateFormatVersion": "2010-09-09",
  "Description": "The AWS CloudFormation template for this Serverless application",
  "Resources": {
    "Cluster1": {
        "Type" : "AWS::EMR::Cluster",
        "Properties" : {
          "Instances" : {
              "CoreInstanceGroup": {
                  "InstanceCount": 3,
                  "InstanceType": "m3g",
              }
          },
          "JobFlowRole" : "EMR_EC2_DefaultRole",
          "Name" : "my cluster",
          "ServiceRole" : "EMR_DefaultRole",
        }
    },
  },
  "Outputs": {
    "ClusterId": {
        "Description": "Cluster info",
        "Value": {
            "Fn::GetAtt": ["Cluster1", "Id"]
        },
    }
  }
}"""


@mock_aws
def test_create_simple_cluster__using_cloudformation():
    region = "us-east-1"
    cf = boto3.client("cloudformation", region_name=region)
    emr = boto3.client("emr", region_name=region)
    cf.create_stack(StackName="teststack", TemplateBody=template)

    # Verify resources
    res = cf.describe_stack_resources(StackName="teststack")["StackResources"][0]
    cluster_id = res["PhysicalResourceId"]
    assert res["LogicalResourceId"] == "Cluster1"
    assert res["ResourceType"] == "AWS::EMR::Cluster"
    assert cluster_id.startswith("j-")

    # Verify outputs
    stack = cf.describe_stacks(StackName="teststack")["Stacks"][0]
    assert {
        "OutputKey": "ClusterId",
        "OutputValue": cluster_id,
        "Description": "Cluster info",
    } in stack["Outputs"]

    # Verify EMR Cluster
    cl = emr.describe_cluster(ClusterId=cluster_id)["Cluster"]
    assert cl["Name"] == "my cluster"
    assert cl["Ec2InstanceAttributes"]["IamInstanceProfile"] == "EMR_EC2_DefaultRole"
    assert cl["ServiceRole"] == "EMR_DefaultRole"
    assert cl["Tags"] == []


template_with_tags = {
    "Resources": {
        "Cluster1": {
            "Type": "AWS::EMR::Cluster",
            "Properties": {
                "Instances": {
                    "CoreInstanceGroup": {
                        "InstanceCount": 3,
                        "InstanceType": "m3g",
                    }
                },
                "JobFlowRole": "EMR_EC2_DefaultRole",
                "Name": "my cluster",
                "ServiceRole": "EMR_DefaultRole",
                "Tags": [
                    {"Key": "k1", "Value": "v1"},
                    {"Key": "k2", "Value": "v2"},
                ],
            },
        },
    },
}


@mock_aws
def test_create_simple_cluster_with_tags():
    region = "us-east-1"
    cf = boto3.client("cloudformation", region_name=region)
    emr = boto3.client("emr", region_name=region)
    cf.create_stack(StackName="teststack", TemplateBody=json.dumps(template_with_tags))

    # Verify resources
    res = cf.describe_stack_resources(StackName="teststack")["StackResources"][0]
    cluster_id = res["PhysicalResourceId"]

    # Verify EMR Cluster
    cl = emr.describe_cluster(ClusterId=cluster_id)["Cluster"]
    assert cl["Tags"] == [{"Key": "k1", "Value": "v1"}, {"Key": "k2", "Value": "v2"}]


template_with_custom_ami = {
    "AWSTemplateFormatVersion": "2010-09-09",
    "Parameters": {
        "SubnetId": {"Type": "String"},
        "InstanceType": {"Type": "String"},
        "TerminationProtected": {"Type": "String", "Default": "false"},
        "ElasticMapReducePrincipal": {"Type": "String"},
        "Ec2Principal": {"Type": "String"},
    },
    "Resources": {
        "cluster": {
            "Type": "AWS::EMR::Cluster",
            "Properties": {
                "CustomAmiId": EXAMPLE_AMI_ID,
                "Instances": {
                    "MasterInstanceGroup": {
                        "InstanceCount": 1,
                        "InstanceType": {"Ref": "InstanceType"},
                        "Market": "ON_DEMAND",
                        "Name": "cfnMaster",
                    },
                    "CoreInstanceGroup": {
                        "InstanceCount": 1,
                        "InstanceType": {"Ref": "InstanceType"},
                        "Market": "ON_DEMAND",
                        "Name": "cfnCore",
                    },
                    "TaskInstanceGroups": [
                        {
                            "InstanceCount": 1,
                            "InstanceType": {"Ref": "InstanceType"},
                            "Market": "ON_DEMAND",
                            "Name": "cfnTask-1",
                        },
                        {
                            "InstanceCount": 1,
                            "InstanceType": {"Ref": "InstanceType"},
                            "Market": "ON_DEMAND",
                            "Name": "cfnTask-2",
                        },
                    ],
                    "TerminationProtected": {"Ref": "TerminationProtected"},
                    "Ec2SubnetId": {"Ref": "SubnetId"},
                },
                "Name": "CFNtest",
                "JobFlowRole": {"Ref": "emrEc2InstanceProfile"},
                "ServiceRole": {"Ref": "emrRole"},
                "ReleaseLabel": "release_2024",
                "VisibleToAllUsers": True,
                "Tags": [{"Key": "key1", "Value": "value1"}],
            },
        },
        "emrRole": {
            "Type": "AWS::IAM::Role",
            "Properties": {
                "AssumeRolePolicyDocument": {
                    "Version": "2008-10-17",
                    "Statement": [
                        {
                            "Sid": "",
                            "Effect": "Allow",
                            "Principal": {
                                "Service": {"Ref": "ElasticMapReducePrincipal"}
                            },
                            "Action": "sts:AssumeRole",
                        }
                    ],
                },
                "Path": "/",
                "ManagedPolicyArns": [
                    "arn:aws:iam::aws:policy/service-role/AmazonElasticMapReduceRole"
                ],
            },
        },
        "emrEc2Role": {
            "Type": "AWS::IAM::Role",
            "Properties": {
                "AssumeRolePolicyDocument": {
                    "Version": "2008-10-17",
                    "Statement": [
                        {
                            "Sid": "",
                            "Effect": "Allow",
                            "Principal": {"Service": {"Ref": "Ec2Principal"}},
                            "Action": "sts:AssumeRole",
                        }
                    ],
                },
                "Path": "/",
                "ManagedPolicyArns": [
                    "arn:aws:iam::aws:policy/service-role/AmazonElasticMapReduceforEC2Role"
                ],
            },
        },
        "emrEc2InstanceProfile": {
            "Type": "AWS::IAM::InstanceProfile",
            "Properties": {"Path": "/", "Roles": [{"Ref": "emrEc2Role"}]},
        },
    },
}


@mock_aws
def test_create_cluster_with_custom_ami():
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    subnet = ec2.create_subnet(VpcId=vpc.id, CidrBlock="10.0.0.0/18")

    params = [
        {"ParameterKey": "InstanceType", "ParameterValue": "i4s"},
        {"ParameterKey": "SubnetId", "ParameterValue": subnet.id},
        {
            "ParameterKey": "ElasticMapReducePrincipal",
            "ParameterValue": "emr_principal",
        },
        {"ParameterKey": "Ec2Principal", "ParameterValue": "ec2_principal"},
    ]

    region = "us-east-1"
    cf = boto3.client("cloudformation", region_name=region)
    emr = boto3.client("emr", region_name=region)
    cf.create_stack(
        StackName="teststack",
        TemplateBody=json.dumps(template_with_custom_ami),
        Parameters=params,
    )

    # Verify resources
    res = cf.describe_stack_resources(StackName="teststack")["StackResources"]
    cluster = [r for r in res if r["ResourceType"] == "AWS::EMR::Cluster"][0]
    cluster_id = cluster["PhysicalResourceId"]

    instance_profile = [
        r for r in res if r["ResourceType"] == "AWS::IAM::InstanceProfile"
    ][0]
    instance_profile_id = instance_profile["PhysicalResourceId"]

    role = [r for r in res if r["ResourceType"] == "AWS::IAM::Role"][0]
    role_id = role["PhysicalResourceId"]

    # Verify EMR Cluster
    cl = emr.describe_cluster(ClusterId=cluster_id)["Cluster"]
    assert cl["Name"] == "CFNtest"
    assert cl["Ec2InstanceAttributes"]["Ec2SubnetId"] == subnet.id
    assert cl["Ec2InstanceAttributes"]["IamInstanceProfile"] == instance_profile_id
    assert cl["ServiceRole"] == role_id
    assert cl["ReleaseLabel"] == "release_2024"
    assert cl["CustomAmiId"] == EXAMPLE_AMI_ID


template_with_root_volume = {
    "AWSTemplateFormatVersion": "2010-09-09",
    "Parameters": {
        "InstanceType": {"Type": "String"},
        "ReleaseLabel": {"Type": "String"},
        "SubnetId": {"Type": "String"},
        "TerminationProtected": {"Type": "String", "Default": "false"},
        "EbsRootVolumeSize": {"Type": "String"},
    },
    "Resources": {
        "cluster": {
            "Type": "AWS::EMR::Cluster",
            "Properties": {
                "EbsRootVolumeSize": {"Ref": "EbsRootVolumeSize"},
                "Instances": {
                    "MasterInstanceGroup": {
                        "InstanceCount": 1,
                        "InstanceType": {"Ref": "InstanceType"},
                        "Market": "ON_DEMAND",
                        "Name": "cfnMaster",
                    },
                    "CoreInstanceGroup": {
                        "InstanceCount": 1,
                        "InstanceType": {"Ref": "InstanceType"},
                        "Market": "ON_DEMAND",
                        "Name": "cfnCore",
                    },
                    "TaskInstanceGroups": [
                        {
                            "InstanceCount": 1,
                            "InstanceType": {"Ref": "InstanceType"},
                            "Market": "ON_DEMAND",
                            "Name": "cfnTask-1",
                        },
                        {
                            "InstanceCount": 1,
                            "InstanceType": {"Ref": "InstanceType"},
                            "Market": "ON_DEMAND",
                            "Name": "cfnTask-2",
                        },
                    ],
                    "TerminationProtected": {"Ref": "TerminationProtected"},
                    "Ec2SubnetId": {"Ref": "SubnetId"},
                },
                "Name": "CFNtest",
                "JobFlowRole": {"Ref": "emrEc2InstanceProfile"},
                "ServiceRole": {"Ref": "emrRole"},
                "ReleaseLabel": {"Ref": "ReleaseLabel"},
                "VisibleToAllUsers": True,
                "Tags": [{"Key": "key1", "Value": "value1"}],
            },
        },
        "emrRole": {
            "Type": "AWS::IAM::Role",
            "Properties": {
                "AssumeRolePolicyDocument": {
                    "Version": "2008-10-17",
                    "Statement": [
                        {
                            "Sid": "",
                            "Effect": "Allow",
                            "Principal": {"Service": "elasticmapreduce.amazonaws.com"},
                            "Action": "sts:AssumeRole",
                        }
                    ],
                },
                "Path": "/",
                "ManagedPolicyArns": [
                    "arn:aws:iam::aws:policy/service-role/AmazonElasticMapReduceRole"
                ],
            },
        },
        "emrEc2Role": {
            "Type": "AWS::IAM::Role",
            "Properties": {
                "AssumeRolePolicyDocument": {
                    "Version": "2008-10-17",
                    "Statement": [
                        {
                            "Sid": "",
                            "Effect": "Allow",
                            "Principal": {"Service": "ec2.amazonaws.com"},
                            "Action": "sts:AssumeRole",
                        }
                    ],
                },
                "Path": "/",
                "ManagedPolicyArns": [
                    "arn:aws:iam::aws:policy/service-role/AmazonElasticMapReduceforEC2Role"
                ],
            },
        },
        "emrEc2InstanceProfile": {
            "Type": "AWS::IAM::InstanceProfile",
            "Properties": {"Path": "/", "Roles": [{"Ref": "emrEc2Role"}]},
        },
    },
}


@mock_aws
def test_create_cluster_with_root_volume():
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    subnet = ec2.create_subnet(VpcId=vpc.id, CidrBlock="10.0.0.0/18")

    params = [
        {"ParameterKey": "InstanceType", "ParameterValue": "i4s"},
        {"ParameterKey": "ReleaseLabel", "ParameterValue": "latest"},
        {"ParameterKey": "EbsRootVolumeSize", "ParameterValue": "15"},
        {"ParameterKey": "SubnetId", "ParameterValue": subnet.id},
        {
            "ParameterKey": "ElasticMapReducePrincipal",
            "ParameterValue": "emr_principal",
        },
        {"ParameterKey": "Ec2Principal", "ParameterValue": "ec2_principal"},
    ]

    region = "us-east-1"
    cf = boto3.client("cloudformation", region_name=region)
    emr = boto3.client("emr", region_name=region)
    cf.create_stack(
        StackName="teststack",
        TemplateBody=json.dumps(template_with_root_volume),
        Parameters=params,
    )

    # Verify resources
    res = cf.describe_stack_resources(StackName="teststack")["StackResources"]
    cluster = [r for r in res if r["ResourceType"] == "AWS::EMR::Cluster"][0]
    cluster_id = cluster["PhysicalResourceId"]

    instance_profile = [
        r for r in res if r["ResourceType"] == "AWS::IAM::InstanceProfile"
    ][0]
    instance_profile_id = instance_profile["PhysicalResourceId"]

    role = [r for r in res if r["ResourceType"] == "AWS::IAM::Role"][0]
    role_id = role["PhysicalResourceId"]

    # Verify EMR Cluster
    cl = emr.describe_cluster(ClusterId=cluster_id)["Cluster"]
    assert cl["Name"] == "CFNtest"
    assert cl["Ec2InstanceAttributes"]["Ec2SubnetId"] == subnet.id
    assert cl["Ec2InstanceAttributes"]["IamInstanceProfile"] == instance_profile_id
    assert cl["ServiceRole"] == role_id


template_with_kerberos_attrs = {
    "AWSTemplateFormatVersion": "2010-09-09",
    "Parameters": {
        "CrossRealmTrustPrincipalPassword": {"Type": "String"},
        "KdcAdminPassword": {"Type": "String"},
        "InstanceType": {"Type": "String"},
        "ReleaseLabel": {"Type": "String"},
        "SubnetId": {"Type": "String"},
    },
    "Resources": {
        "cluster": {
            "Type": "AWS::EMR::Cluster",
            "Properties": {
                "Instances": {
                    "MasterInstanceGroup": {
                        "InstanceCount": 1,
                        "InstanceType": {"Ref": "InstanceType"},
                        "Market": "ON_DEMAND",
                        "Name": "cfnMaster",
                    },
                    "CoreInstanceGroup": {
                        "InstanceCount": 1,
                        "InstanceType": {"Ref": "InstanceType"},
                        "Market": "ON_DEMAND",
                        "Name": "cfnCore",
                    },
                    "TaskInstanceGroups": [
                        {
                            "InstanceCount": 1,
                            "InstanceType": {"Ref": "InstanceType"},
                            "Market": "ON_DEMAND",
                            "Name": "cfnTask-1",
                        },
                        {
                            "InstanceCount": 1,
                            "InstanceType": {"Ref": "InstanceType"},
                            "Market": "ON_DEMAND",
                            "Name": "cfnTask-2",
                        },
                    ],
                    "Ec2SubnetId": {"Ref": "SubnetId"},
                },
                "Name": "CFNtest",
                "JobFlowRole": {"Ref": "emrEc2InstanceProfile"},
                "KerberosAttributes": {
                    "CrossRealmTrustPrincipalPassword": {
                        "Ref": "CrossRealmTrustPrincipalPassword"
                    },
                    "KdcAdminPassword": {"Ref": "KdcAdminPassword"},
                    "Realm": "EC2.INTERNAL",
                },
                "ServiceRole": {"Ref": "emrRole"},
                "ReleaseLabel": {"Ref": "ReleaseLabel"},
                "SecurityConfiguration": {"Ref": "securityConfiguration"},
                "VisibleToAllUsers": True,
                "Tags": [{"Key": "key1", "Value": "value1"}],
            },
        },
        "key": {
            "Type": "AWS::KMS::Key",
            "Properties": {
                "KeyPolicy": {
                    "Version": "2012-10-17",
                    "Id": "key-default-1",
                    "Statement": [
                        {
                            "Sid": "Enable IAM User Permissions",
                            "Effect": "Allow",
                            "Principal": {"AWS": {"Fn::GetAtt": ["emrEc2Role", "Arn"]}},
                            "Action": "kms:*",
                            "Resource": "*",
                        },
                        {
                            "Sid": "Enable IAM User Permissions",
                            "Effect": "Allow",
                            "Principal": {
                                "AWS": {
                                    "Fn::Join": [
                                        "",
                                        [
                                            "arn:aws:iam::",
                                            {"Ref": "AWS::AccountId"},
                                            ":root",
                                        ],
                                    ]
                                }
                            },
                            "Action": "kms:*",
                            "Resource": "*",
                        },
                    ],
                }
            },
        },
        "securityConfiguration": {
            "Type": "AWS::EMR::SecurityConfiguration",
            "Properties": {
                "Name": "mysecconfig",
                "SecurityConfiguration": {
                    "AuthenticationConfiguration": {
                        "KerberosConfiguration": {
                            "Provider": "ClusterDedicatedKdc",
                            "ClusterDedicatedKdcConfiguration": {
                                "TicketLifetimeInHours": 24,
                                "CrossRealmTrustConfiguration": {
                                    "Realm": "AD.DOMAIN.COM",
                                    "Domain": "ad.domain.com",
                                    "AdminServer": "ad.domain.com",
                                    "KdcServer": "ad.domain.com",
                                },
                            },
                        }
                    }
                },
            },
        },
        "emrRole": {
            "Type": "AWS::IAM::Role",
            "Properties": {
                "AssumeRolePolicyDocument": {
                    "Version": "2008-10-17",
                    "Statement": [
                        {
                            "Sid": "",
                            "Effect": "Allow",
                            "Principal": {"Service": "elasticmapreduce.amazonaws.com"},
                            "Action": "sts:AssumeRole",
                        }
                    ],
                },
                "Path": "/",
                "ManagedPolicyArns": [
                    "arn:aws:iam::aws:policy/service-role/AmazonElasticMapReduceRole"
                ],
            },
        },
        "emrEc2Role": {
            "Type": "AWS::IAM::Role",
            "Properties": {
                "AssumeRolePolicyDocument": {
                    "Version": "2008-10-17",
                    "Statement": [
                        {
                            "Sid": "",
                            "Effect": "Allow",
                            "Principal": {"Service": "ec2.amazonaws.com"},
                            "Action": "sts:AssumeRole",
                        }
                    ],
                },
                "Path": "/",
                "ManagedPolicyArns": [
                    "arn:aws:iam::aws:policy/service-role/AmazonElasticMapReduceforEC2Role"
                ],
            },
        },
        "emrEc2InstanceProfile": {
            "Type": "AWS::IAM::InstanceProfile",
            "Properties": {"Path": "/", "Roles": [{"Ref": "emrEc2Role"}]},
        },
    },
    "Outputs": {"keyArn": {"Value": {"Fn::GetAtt": ["key", "Arn"]}}},
}


@mock_aws
def test_create_cluster_with_kerberos_attrs():
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    subnet = ec2.create_subnet(VpcId=vpc.id, CidrBlock="10.0.0.0/18")

    params = [
        {"ParameterKey": "CrossRealmTrustPrincipalPassword", "ParameterValue": "p2ss"},
        {"ParameterKey": "KdcAdminPassword", "ParameterValue": "adminp2ss"},
        {"ParameterKey": "InstanceType", "ParameterValue": "i4s"},
        {"ParameterKey": "ReleaseLabel", "ParameterValue": "latest"},
        {"ParameterKey": "EbsRootVolumeSize", "ParameterValue": "15"},
        {"ParameterKey": "SubnetId", "ParameterValue": subnet.id},
        {
            "ParameterKey": "ElasticMapReducePrincipal",
            "ParameterValue": "emr_principal",
        },
        {"ParameterKey": "Ec2Principal", "ParameterValue": "ec2_principal"},
    ]

    region = "us-east-1"
    cf = boto3.client("cloudformation", region_name=region)
    emr = boto3.client("emr", region_name=region)
    cf.create_stack(
        StackName="teststack",
        TemplateBody=json.dumps(template_with_kerberos_attrs),
        Parameters=params,
    )

    # Verify resources
    res = cf.describe_stack_resources(StackName="teststack")["StackResources"]
    cluster = [r for r in res if r["ResourceType"] == "AWS::EMR::Cluster"][0]
    cluster_id = cluster["PhysicalResourceId"]

    instance_profile = [
        r for r in res if r["ResourceType"] == "AWS::IAM::InstanceProfile"
    ][0]
    instance_profile_id = instance_profile["PhysicalResourceId"]

    role = [r for r in res if r["ResourceType"] == "AWS::IAM::Role"][0]
    role_id = role["PhysicalResourceId"]

    # Verify EMR Cluster
    cl = emr.describe_cluster(ClusterId=cluster_id)["Cluster"]
    assert cl["Name"] == "CFNtest"
    assert cl["Ec2InstanceAttributes"]["Ec2SubnetId"] == subnet.id
    assert cl["Ec2InstanceAttributes"]["IamInstanceProfile"] == instance_profile_id
    assert cl["ServiceRole"] == role_id

    kerberos = cl["KerberosAttributes"]
    assert kerberos == {
        "Realm": "EC2.INTERNAL",
        "KdcAdminPassword": "adminp2ss",
        "CrossRealmTrustPrincipalPassword": "p2ss",
    }

    # Verify everything can be deleted
    cf.delete_stack(StackName="teststack")

    cl = emr.describe_cluster(ClusterId=cluster_id)["Cluster"]
    assert cl["Status"]["State"] == "TERMINATED"

    with pytest.raises(ClientError) as exc:
        emr.describe_security_configuration(Name="mysecconfig")
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidRequestException"


template_with_simple_instance_group_config = {
    "Resources": {
        "Cluster1": {
            "Type": "AWS::EMR::Cluster",
            "Properties": {
                "Instances": {
                    "CoreInstanceGroup": {
                        "InstanceCount": 3,
                        "InstanceType": "m3g",
                    }
                },
                "JobFlowRole": "EMR_EC2_DefaultRole",
                "Name": "my cluster",
                "ServiceRole": "EMR_DefaultRole",
            },
        },
        "TestInstanceGroupConfig": {
            "Type": "AWS::EMR::InstanceGroupConfig",
            "Properties": {
                "InstanceCount": 2,
                "InstanceType": "m3.xlarge",
                "InstanceRole": "TASK",
                "Market": "ON_DEMAND",
                "Name": "cfnTask2",
                "JobFlowId": {"Ref": "Cluster1"},
            },
        },
    },
}


@mock_aws
def test_create_simple_instance_group():
    region = "us-east-1"
    cf = boto3.client("cloudformation", region_name=region)
    emr = boto3.client("emr", region_name=region)
    cf.create_stack(
        StackName="teststack",
        TemplateBody=json.dumps(template_with_simple_instance_group_config),
    )

    # Verify resources
    res = cf.describe_stack_resources(StackName="teststack")["StackResources"][0]
    cluster_id = res["PhysicalResourceId"]

    ig = emr.list_instance_groups(ClusterId=cluster_id)["InstanceGroups"][0]
    assert ig["Name"] == "cfnTask2"
    assert ig["Market"] == "ON_DEMAND"
    assert ig["InstanceGroupType"] == "TASK"
    assert ig["InstanceType"] == "m3.xlarge"


template_with_advanced_instance_group_config = {
    "Resources": {
        "Cluster1": {
            "Type": "AWS::EMR::Cluster",
            "Properties": {
                "Instances": {
                    "CoreInstanceGroup": {
                        "InstanceCount": 3,
                        "InstanceType": "m3g",
                    }
                },
                "JobFlowRole": "EMR_EC2_DefaultRole",
                "Name": "my cluster",
                "ServiceRole": "EMR_DefaultRole",
            },
        },
        "TestInstanceGroupConfig": {
            "Type": "AWS::EMR::InstanceGroupConfig",
            "Properties": {
                "InstanceCount": 1,
                "InstanceType": "m4.large",
                "InstanceRole": "TASK",
                "Market": "ON_DEMAND",
                "Name": "cfnTask3",
                "JobFlowId": {"Ref": "Cluster1"},
                "EbsConfiguration": {
                    "EbsOptimized": True,
                    "EbsBlockDeviceConfigs": [
                        {
                            "VolumesPerInstance": 2,
                            "VolumeSpecification": {
                                "Iops": 10,
                                "SizeInGB": 50,
                                "Throughput": 100,
                                "VolumeType": "gp3",
                            },
                        }
                    ],
                },
                "AutoScalingPolicy": {
                    "Constraints": {"MinCapacity": 1, "MaxCapacity": 4},
                    "Rules": [
                        {
                            "Name": "Scale-out",
                            "Description": "Scale-out policy",
                            "Action": {
                                "SimpleScalingPolicyConfiguration": {
                                    "AdjustmentType": "CHANGE_IN_CAPACITY",
                                    "ScalingAdjustment": 1,
                                    "CoolDown": 300,
                                }
                            },
                            "Trigger": {
                                "CloudWatchAlarmDefinition": {
                                    "Dimensions": [
                                        {
                                            "Key": "JobFlowId",
                                            "Value": "${emr.clusterId}",
                                        }
                                    ],
                                    "EvaluationPeriods": 1,
                                    "Namespace": "AWS/ElasticMapReduce",
                                    "Period": 300,
                                    "ComparisonOperator": "LESS_THAN",
                                    "Statistic": "AVERAGE",
                                    "Threshold": 15,
                                    "Unit": "PERCENT",
                                    "MetricName": "YARNMemoryAvailablePercentage",
                                }
                            },
                        },
                        {
                            "Name": "Scale-in",
                            "Description": "Scale-in policy",
                            "Action": {
                                "SimpleScalingPolicyConfiguration": {
                                    "AdjustmentType": "CHANGE_IN_CAPACITY",
                                    "ScalingAdjustment": -1,
                                    "CoolDown": 300,
                                }
                            },
                            "Trigger": {
                                "CloudWatchAlarmDefinition": {
                                    "Dimensions": [
                                        {
                                            "Key": "JobFlowId",
                                            "Value": "${emr.clusterId}",
                                        }
                                    ],
                                    "EvaluationPeriods": 1,
                                    "Namespace": "AWS/ElasticMapReduce",
                                    "Period": 300,
                                    "ComparisonOperator": "GREATER_THAN",
                                    "Statistic": "AVERAGE",
                                    "Threshold": 75,
                                    "Unit": "PERCENT",
                                    "MetricName": "YARNMemoryAvailablePercentage",
                                }
                            },
                        },
                    ],
                },
            },
        },
    },
}


@mock_aws
def test_create_advanced_instance_group():
    region = "us-east-1"
    cf = boto3.client("cloudformation", region_name=region)
    emr = boto3.client("emr", region_name=region)
    cf.create_stack(
        StackName="teststack",
        TemplateBody=json.dumps(template_with_advanced_instance_group_config),
    )

    # Verify resources
    res = cf.describe_stack_resources(StackName="teststack")["StackResources"][0]
    cluster_id = res["PhysicalResourceId"]

    ig = emr.list_instance_groups(ClusterId=cluster_id)["InstanceGroups"][0]
    assert ig["Name"] == "cfnTask3"
    assert ig["Market"] == "ON_DEMAND"
    assert ig["InstanceGroupType"] == "TASK"
    assert ig["InstanceType"] == "m4.large"

    as_policy = ig["AutoScalingPolicy"]
    assert as_policy["Status"] == {"State": "ATTACHED"}
    assert as_policy["Constraints"] == {"MinCapacity": 1, "MaxCapacity": 4}

    ebs = ig["EbsBlockDevices"]
    assert ebs[0]["VolumeSpecification"] == {
        "VolumeType": "gp3",
        "Iops": 10,
        "SizeInGB": 50,
    }



import json

import boto3

from moto import mock_aws
from tests.test_cloudformation.fixtures import redshift


@mock_aws
def test_redshift_stack():
    redshift_template_json = json.dumps(redshift.template)

    ec2 = boto3.client("ec2", region_name="us-west-2")
    cf = boto3.client("cloudformation", region_name="us-west-2")
    cf.create_stack(
        StackName="redshift_stack",
        TemplateBody=redshift_template_json,
        Parameters=[
            {"ParameterKey": "DatabaseName", "ParameterValue": "mydb"},
            {"ParameterKey": "ClusterType", "ParameterValue": "multi-node"},
            {"ParameterKey": "NumberOfNodes", "ParameterValue": "2"},
            {"ParameterKey": "NodeType", "ParameterValue": "dw1.xlarge"},
            {"ParameterKey": "MasterUsername", "ParameterValue": "myuser"},
            {"ParameterKey": "MasterUserPassword", "ParameterValue": "mypass"},
            {"ParameterKey": "InboundTraffic", "ParameterValue": "10.0.0.1/16"},
            {"ParameterKey": "PortNumber", "ParameterValue": "5439"},
        ],
    )

    redshift_conn = boto3.client("redshift", region_name="us-west-2")

    cluster_res = redshift_conn.describe_clusters()
    clusters = cluster_res["Clusters"]
    assert len(clusters) == 1
    cluster = clusters[0]
    assert cluster["DBName"] == "mydb"
    assert cluster["NumberOfNodes"] == 2
    assert cluster["NodeType"] == "dw1.xlarge"
    assert cluster["MasterUsername"] == "myuser"
    assert cluster["Endpoint"]["Port"] == 5439
    assert len(cluster["VpcSecurityGroups"]) == 1
    security_group_id = cluster["VpcSecurityGroups"][0]["VpcSecurityGroupId"]

    groups = ec2.describe_security_groups(GroupIds=[security_group_id])[
        "SecurityGroups"
    ]
    assert len(groups) == 1
    group = groups[0]
    assert len(group["IpPermissions"]) == 1
    assert group["IpPermissions"][0]["IpRanges"][0]["CidrIp"] == "10.0.0.1/16"



from moto.opensearch.responses import OpenSearchServiceResponse

from .responses import ElasticsearchServiceResponse

url_bases = [
    r"https?://es\.(.+)\.amazonaws\.com",
]


url_paths = {
    "{0}/2015-01-01/domain$": ElasticsearchServiceResponse.list_domains,
    "{0}/2015-01-01/es/domain$": ElasticsearchServiceResponse.domains,
    "{0}/2015-01-01/es/domain/(?P<domainname>[^/]+)": ElasticsearchServiceResponse.domain,
    "{0}/2021-01-01/domain$": OpenSearchServiceResponse.dispatch,
    "{0}/2021-01-01/opensearch/compatibleVersions": OpenSearchServiceResponse.dispatch,
    "{0}/2021-01-01/opensearch/domain": OpenSearchServiceResponse.dispatch,
    "{0}/2021-01-01/opensearch/domain/(?P<domainname>[^/]+)": OpenSearchServiceResponse.dispatch,
    "{0}/2021-01-01/opensearch/domain/(?P<domainname>[^/]+)/config": OpenSearchServiceResponse.dispatch,
    "{0}/2021-01-01/tags/?": OpenSearchServiceResponse.dispatch,
    "{0}/2021-01-01/tags-removal/": OpenSearchServiceResponse.dispatch,
}



import abc
from typing import Final

from moto.stepfunctions.parser.api import ExecutionFailedEventDetails, HistoryEventType
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
    FailureEventException,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name import (
    StatesErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name_type import (
    StatesErrorNameType,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails
from moto.stepfunctions.parser.asl.utils.encoding import to_json_str
from moto.stepfunctions.parser.asl.utils.json_path import JSONPathUtils

TOLERATED_FAILURE_COUNT_MIN: Final[int] = 0
TOLERATED_FAILURE_COUNT_DEFAULT: Final[int] = 0
TOLERATED_FAILURE_PERCENTAGE_MIN: Final[float] = 0.0
TOLERATED_FAILURE_PERCENTAGE_DEFAULT: Final[float] = 0.0
TOLERATED_FAILURE_PERCENTAGE_MAX: Final[float] = 100.0


class ToleratedFailureCountDecl(EvalComponent, abc.ABC):
    @abc.abstractmethod
    def _eval_tolerated_failure_count(self, env: Environment) -> int: ...

    def _eval_body(self, env: Environment) -> None:
        tolerated_failure_count = self._eval_tolerated_failure_count(env=env)
        env.stack.append(tolerated_failure_count)


class ToleratedFailureCount(ToleratedFailureCountDecl):
    tolerated_failure_count: Final[int]

    def __init__(self, tolerated_failure_count: int = TOLERATED_FAILURE_COUNT_DEFAULT):
        self.tolerated_failure_count = tolerated_failure_count

    def _eval_tolerated_failure_count(self, env: Environment) -> int:
        return self.tolerated_failure_count


class ToleratedFailureCountPath(ToleratedFailureCountDecl):
    tolerated_failure_count_path: Final[str]

    def __init__(self, tolerated_failure_count_path: str):
        self.tolerated_failure_count_path = tolerated_failure_count_path

    def _eval_tolerated_failure_count(self, env: Environment) -> int:
        inp = env.stack[-1]
        tolerated_failure_count = JSONPathUtils.extract_json(
            self.tolerated_failure_count_path, inp
        )

        error_cause = None
        if not isinstance(tolerated_failure_count, int):
            value_str = (
                to_json_str(tolerated_failure_count)
                if not isinstance(tolerated_failure_count, str)
                else tolerated_failure_count
            )
            error_cause = (
                f'The ToleratedFailureCountPath field refers to value "{value_str}" '
                f"which is not a valid integer: {self.tolerated_failure_count_path}"
            )

        elif tolerated_failure_count < TOLERATED_FAILURE_COUNT_MIN:
            error_cause = "ToleratedFailureCount cannot be negative."

        if error_cause is not None:
            raise FailureEventException(
                failure_event=FailureEvent(
                    env=env,
                    error_name=StatesErrorName(typ=StatesErrorNameType.StatesRuntime),
                    event_type=HistoryEventType.ExecutionFailed,
                    event_details=EventDetails(
                        executionFailedEventDetails=ExecutionFailedEventDetails(
                            error=StatesErrorNameType.StatesRuntime.to_name(),
                            cause=error_cause,
                        )
                    ),
                )
            )

        return tolerated_failure_count


class ToleratedFailurePercentageDecl(EvalComponent, abc.ABC):
    @abc.abstractmethod
    def _eval_tolerated_failure_percentage(self, env: Environment) -> float: ...

    def _eval_body(self, env: Environment) -> None:
        tolerated_failure_percentage = self._eval_tolerated_failure_percentage(env=env)
        env.stack.append(tolerated_failure_percentage)


class ToleratedFailurePercentage(ToleratedFailurePercentageDecl):
    tolerated_failure_percentage: Final[float]

    def __init__(
        self, tolerated_failure_percentage: float = TOLERATED_FAILURE_PERCENTAGE_DEFAULT
    ):
        self.tolerated_failure_percentage = tolerated_failure_percentage

    def _eval_tolerated_failure_percentage(self, env: Environment) -> float:
        return self.tolerated_failure_percentage


class ToleratedFailurePercentagePath(ToleratedFailurePercentageDecl):
    tolerate_failure_percentage_path: Final[str]

    def __init__(self, tolerate_failure_percentage_path: str):
        self.tolerate_failure_percentage_path = tolerate_failure_percentage_path

    def _eval_tolerated_failure_percentage(self, env: Environment) -> float:
        inp = env.stack[-1]
        tolerated_failure_percentage = JSONPathUtils.extract_json(
            self.tolerate_failure_percentage_path, inp
        )

        if isinstance(tolerated_failure_percentage, int):
            tolerated_failure_percentage = float(tolerated_failure_percentage)

        error_cause = None
        if not isinstance(tolerated_failure_percentage, float):
            value_str = (
                to_json_str(tolerated_failure_percentage)
                if not isinstance(tolerated_failure_percentage, str)
                else tolerated_failure_percentage
            )
            error_cause = (
                f'The ToleratedFailurePercentagePath field refers to value "{value_str}" '
                f"which is not a valid float: {self.tolerate_failure_percentage_path}"
            )
        elif (
            not TOLERATED_FAILURE_PERCENTAGE_MIN
            <= tolerated_failure_percentage
            <= TOLERATED_FAILURE_PERCENTAGE_MAX
        ):
            error_cause = "ToleratedFailurePercentage must be between 0 and 100."

        if error_cause is not None:
            raise FailureEventException(
                failure_event=FailureEvent(
                    env=env,
                    error_name=StatesErrorName(typ=StatesErrorNameType.StatesRuntime),
                    event_type=HistoryEventType.ExecutionFailed,
                    event_details=error_cause,
                )
            )

        return tolerated_failure_percentage



import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


@mock_aws
def test_list_aws_default_service_quotas():
    client = boto3.client("service-quotas", region_name="eu-west-1")
    resp = client.list_aws_default_service_quotas(ServiceCode="vpc")

    assert len(resp["Quotas"]) == 25

    assert {
        "Adjustable": True,
        "GlobalQuota": False,
        "QuotaArn": "arn:aws:servicequotas:eu-west-1::vpc/L-2AFB9258",
        "QuotaCode": "L-2AFB9258",
        "QuotaName": "Security groups per network interface",
        "ServiceCode": "vpc",
        "ServiceName": "Amazon Virtual Private Cloud (Amazon VPC)",
        "Unit": "None",
        "Value": 5.0,
    } in resp["Quotas"]
    assert {
        "Adjustable": True,
        "GlobalQuota": False,
        "QuotaArn": "arn:aws:servicequotas:eu-west-1::vpc/L-F678F1CE",
        "QuotaCode": "L-F678F1CE",
        "QuotaName": "VPCs per Region",
        "ServiceCode": "vpc",
        "ServiceName": "Amazon Virtual Private Cloud (Amazon VPC)",
        "Unit": "None",
        "Value": 5.0,
    } in resp["Quotas"]


@mock_aws
def test_list_defaults_for_unknown_service():
    client = boto3.client("service-quotas", "us-east-1")

    with pytest.raises(ClientError) as exc:
        client.list_aws_default_service_quotas(ServiceCode="unknown")
    err = exc.value.response["Error"]
    assert err["Code"] == "NoSuchResourceException"
    assert err["Message"] == (
        "This service is not available in the current Region. Choose a "
        "different Region or a different service."
    )


@mock_aws
def test_get_service_quota():
    client = boto3.client("service-quotas", region_name="us-east-2")
    quotas = client.list_aws_default_service_quotas(ServiceCode="vpc")["Quotas"]

    for quota in quotas:
        resp = client.get_service_quota(ServiceCode="vpc", QuotaCode=quota["QuotaCode"])
        assert quota == resp["Quota"]


@mock_aws
def test_get_unknown_service_quota():
    client = boto3.client("service-quotas", region_name="us-east-2")

    with pytest.raises(ClientError) as exc:
        client.get_service_quota(ServiceCode="vpc", QuotaCode="unknown")
    err = exc.value.response["Error"]
    assert err["Code"] == "NoSuchResourceException"



"""Exceptions raised by the backup service."""

from moto.core.exceptions import JsonRESTError


class BackupClientError(JsonRESTError):
    code = 400


class AlreadyExistsException(BackupClientError):
    def __init__(self, msg: str):
        super().__init__("AlreadyExistsException", f"{msg}")


class ResourceNotFoundException(JsonRESTError):
    def __init__(self, msg: str):
        super().__init__("ResourceNotFoundException", f"{msg}")



import json
from uuid import uuid4

import boto3

from moto import mock_aws
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID
from tests import EXAMPLE_AMI_ID


@mock_aws
def test_flow_logs_by_cloudformation():
    s3 = boto3.resource("s3", region_name="us-west-1")
    client = boto3.client("ec2", region_name="us-west-1")
    cf_client = boto3.client("cloudformation", "us-west-1")

    vpc = client.create_vpc(CidrBlock="10.0.0.0/16")["Vpc"]

    bucket_name = str(uuid4())
    bucket = s3.create_bucket(
        Bucket=bucket_name,
        CreateBucketConfiguration={"LocationConstraint": "us-west-1"},
    )

    flow_log_template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "Template for VPC Flow Logs creation.",
        "Resources": {
            "TestFlowLogs": {
                "Type": "AWS::EC2::FlowLog",
                "Properties": {
                    "ResourceType": "VPC",
                    "ResourceId": vpc["VpcId"],
                    "TrafficType": "ALL",
                    "LogDestinationType": "s3",
                    "LogDestination": "arn:aws:s3:::" + bucket.name,
                    "MaxAggregationInterval": "60",
                    "Tags": [{"Key": "foo", "Value": "bar"}],
                },
            }
        },
    }
    flow_log_template_json = json.dumps(flow_log_template)

    stack_name = str(uuid4())
    cf_client.create_stack(StackName=stack_name, TemplateBody=flow_log_template_json)

    flow_logs = client.describe_flow_logs(
        Filters=[{"Name": "resource-id", "Values": [vpc["VpcId"]]}]
    )["FlowLogs"]
    assert len(flow_logs) == 1
    assert flow_logs[0]["ResourceId"] == vpc["VpcId"]
    assert flow_logs[0]["LogDestination"] == "arn:aws:s3:::" + bucket.name
    assert flow_logs[0]["MaxAggregationInterval"] == 60


@mock_aws
def test_cloudformation():
    dummy_template_json = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Resources": {
            "InstanceProfile": {
                "Type": "AWS::IAM::InstanceProfile",
                "Properties": {"Path": "/", "Roles": []},
            },
            "Ec2Instance": {
                "Type": "AWS::EC2::Instance",
                "Properties": {
                    "IamInstanceProfile": {"Ref": "InstanceProfile"},
                    "KeyName": "mykey1",
                    "ImageId": EXAMPLE_AMI_ID,
                },
            },
        },
    }

    client = boto3.client("ec2", region_name="us-east-1")
    cf_conn = boto3.client("cloudformation", region_name="us-east-1")
    stack_name = str(uuid4())
    cf_conn.create_stack(
        StackName=stack_name, TemplateBody=json.dumps(dummy_template_json)
    )

    resources = cf_conn.list_stack_resources(StackName=stack_name)[
        "StackResourceSummaries"
    ]
    iam_id = resources[0]["PhysicalResourceId"]
    iam_ip_arn = f"arn:aws:iam::{ACCOUNT_ID}:instance-profile/{iam_id}"

    all_assocs = client.describe_iam_instance_profile_associations()[
        "IamInstanceProfileAssociations"
    ]
    our_assoc = [a for a in all_assocs if a["IamInstanceProfile"]["Arn"] == iam_ip_arn]
    assert stack_name in our_assoc[0]["IamInstanceProfile"]["Arn"]
    our_assoc_id = our_assoc[0]["AssociationId"]

    cf_conn.delete_stack(StackName=stack_name)
    associations = client.describe_iam_instance_profile_associations()[
        "IamInstanceProfileAssociations"
    ]
    assert our_assoc_id not in [a["AssociationId"] for a in associations]



from ._base_response import EC2BaseResponse


class CustomerGateways(EC2BaseResponse):
    def create_customer_gateway(self) -> str:
        gateway_type = self._get_param("Type")
        ip_address = self._get_param("IpAddress") or self._get_param("PublicIp")
        bgp_asn = self._get_param("BgpAsn")
        tags = self._parse_tag_specification().get("customer-gateway", {})
        customer_gateway = self.ec2_backend.create_customer_gateway(
            gateway_type, ip_address=ip_address, bgp_asn=bgp_asn, tags=tags
        )
        template = self.response_template(CREATE_CUSTOMER_GATEWAY_RESPONSE)
        return template.render(customer_gateway=customer_gateway)

    def delete_customer_gateway(self) -> str:
        customer_gateway_id = self._get_param("CustomerGatewayId")
        self.ec2_backend.delete_customer_gateway(customer_gateway_id)
        template = self.response_template(DELETE_CUSTOMER_GATEWAY_RESPONSE)
        return template.render(delete_status="true")

    def describe_customer_gateways(self) -> str:
        self.error_on_dryrun()
        filters = self._filters_from_querystring()
        customer_gateway_ids = self._get_multi_param("CustomerGatewayId")
        customer_gateways = self.ec2_backend.describe_customer_gateways(
            filters, customer_gateway_ids
        )
        template = self.response_template(DESCRIBE_CUSTOMER_GATEWAYS_RESPONSE)
        return template.render(customer_gateways=customer_gateways)


CREATE_CUSTOMER_GATEWAY_RESPONSE = """
<CreateCustomerGatewayResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
   <requestId>7a62c49f-347e-4fc4-9331-6e8eEXAMPLE</requestId>
   <customerGateway>
      <customerGatewayId>{{ customer_gateway.id }}</customerGatewayId>
      <state>{{ customer_gateway.state }}</state>
      <type>{{ customer_gateway.type }}</type>
      <ipAddress>{{ customer_gateway.ip_address }}</ipAddress>
      <bgpAsn>{{ customer_gateway.bgp_asn }}</bgpAsn>
      <tagSet>
        {% for tag in customer_gateway.get_tags() %}
          <item>
            <key>{{ tag.key }}</key>
              <value>{{ tag.value }}</value>
          </item>
        {% endfor %}
      </tagSet>
   </customerGateway>
</CreateCustomerGatewayResponse>"""

DELETE_CUSTOMER_GATEWAY_RESPONSE = """
<DeleteCustomerGatewayResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
   <requestId>7a62c49f-347e-4fc4-9331-6e8eEXAMPLE</requestId>
   <return>{{ delete_status }}</return>
</DeleteCustomerGatewayResponse>"""

DESCRIBE_CUSTOMER_GATEWAYS_RESPONSE = """
<DescribeCustomerGatewaysResponse xmlns="http://ec2.amazonaws.com/doc/2014-10- 01/">
  <requestId>7a62c49f-347e-4fc4-9331-6e8eEXAMPLE</requestId>
  <customerGatewaySet>
  {% for customer_gateway in customer_gateways %}
    <item>
       <customerGatewayId>{{ customer_gateway.id }}</customerGatewayId>
       <state>{{ customer_gateway.state }}</state>
       <type>{{ customer_gateway.type }}</type>
       <ipAddress>{{ customer_gateway.ip_address }}</ipAddress>
       <bgpAsn>{{ customer_gateway.bgp_asn }}</bgpAsn>
       <tagSet>
        {% for tag in customer_gateway.get_tags() %}
          <item>
            <key>{{ tag.key }}</key>
            <value>{{ tag.value }}</value>
          </item>
        {% endfor %}
       </tagSet>
    </item>
  {% endfor %}
  </customerGatewaySet>
</DescribeCustomerGatewaysResponse>"""



import json
from uuid import uuid4

import boto3

from moto import mock_aws

from .test_batch_jobs import _get_clients, _setup

# Copy of test_batch/test_batch_cloudformation
# Except that we verify this behaviour still works without docker


DEFAULT_REGION = "eu-central-1"


@mock_aws(config={"batch": {"use_docker": False}})
def test_create_env_cf() -> None:
    ec2_client, iam_client, _, _, _ = _get_clients()
    _, subnet_id, sg_id, iam_arn = _setup(ec2_client, iam_client)

    create_environment_template = {
        "Resources": {
            "ComputeEnvironment": {
                "Type": "AWS::Batch::ComputeEnvironment",
                "Properties": {
                    "Type": "MANAGED",
                    "ComputeResources": {
                        "Type": "EC2",
                        "MinvCpus": 0,
                        "DesiredvCpus": 0,
                        "MaxvCpus": 64,
                        "InstanceTypes": ["optimal"],
                        "Subnets": [subnet_id],
                        "SecurityGroupIds": [sg_id],
                        "InstanceRole": iam_arn.replace("role", "instance-profile"),
                    },
                    "ServiceRole": iam_arn,
                },
            }
        }
    }
    cf_json = json.dumps(create_environment_template)

    cf_conn = boto3.client("cloudformation", DEFAULT_REGION)
    stack_name = str(uuid4())[0:6]
    stack_id = cf_conn.create_stack(StackName=stack_name, TemplateBody=cf_json)[
        "StackId"
    ]

    stack_resources = cf_conn.list_stack_resources(StackName=stack_id)
    summary = stack_resources["StackResourceSummaries"][0]

    assert summary["ResourceStatus"] == "CREATE_COMPLETE"
    # Spot checks on the ARN
    assert "arn:aws:batch:" in summary["PhysicalResourceId"]
    assert stack_name in summary["PhysicalResourceId"]


@mock_aws(config={"batch": {"use_docker": False}})
def test_create_job_queue_cf() -> None:
    ec2_client, iam_client, _, _, _ = _get_clients()
    _, subnet_id, sg_id, iam_arn = _setup(ec2_client, iam_client)

    create_environment_template = {
        "Resources": {
            "ComputeEnvironment": {
                "Type": "AWS::Batch::ComputeEnvironment",
                "Properties": {
                    "Type": "MANAGED",
                    "ComputeResources": {
                        "Type": "EC2",
                        "MinvCpus": 0,
                        "DesiredvCpus": 0,
                        "MaxvCpus": 64,
                        "InstanceTypes": ["optimal"],
                        "Subnets": [subnet_id],
                        "SecurityGroupIds": [sg_id],
                        "InstanceRole": iam_arn.replace("role", "instance-profile"),
                    },
                    "ServiceRole": iam_arn,
                },
            },
            "JobQueue": {
                "Type": "AWS::Batch::JobQueue",
                "Properties": {
                    "Priority": 1,
                    "ComputeEnvironmentOrder": [
                        {
                            "Order": 1,
                            "ComputeEnvironment": {"Ref": "ComputeEnvironment"},
                        }
                    ],
                },
            },
        }
    }
    cf_json = json.dumps(create_environment_template)

    cf_conn = boto3.client("cloudformation", DEFAULT_REGION)
    stack_name = str(uuid4())[0:6]
    stack_id = cf_conn.create_stack(StackName=stack_name, TemplateBody=cf_json)[
        "StackId"
    ]

    stack_resources = cf_conn.list_stack_resources(StackName=stack_id)
    assert len(stack_resources["StackResourceSummaries"]) == 2

    job_queue_resource = list(
        filter(
            lambda item: item["ResourceType"] == "AWS::Batch::JobQueue",
            stack_resources["StackResourceSummaries"],
        )
    )[0]

    assert job_queue_resource["ResourceStatus"] == "CREATE_COMPLETE"
    # Spot checks on the ARN
    assert job_queue_resource["PhysicalResourceId"].startswith("arn:aws:batch:")
    assert stack_name in job_queue_resource["PhysicalResourceId"]
    assert "job-queue/" in job_queue_resource["PhysicalResourceId"]


@mock_aws(config={"batch": {"use_docker": False}})
def test_create_job_def_cf() -> None:
    ec2_client, iam_client, _, _, _ = _get_clients()
    _, subnet_id, sg_id, iam_arn = _setup(ec2_client, iam_client)

    create_environment_template = {
        "Resources": {
            "ComputeEnvironment": {
                "Type": "AWS::Batch::ComputeEnvironment",
                "Properties": {
                    "Type": "MANAGED",
                    "ComputeResources": {
                        "Type": "EC2",
                        "MinvCpus": 0,
                        "DesiredvCpus": 0,
                        "MaxvCpus": 64,
                        "InstanceTypes": ["optimal"],
                        "Subnets": [subnet_id],
                        "SecurityGroupIds": [sg_id],
                        "InstanceRole": iam_arn.replace("role", "instance-profile"),
                    },
                    "ServiceRole": iam_arn,
                },
            },
            "JobQueue": {
                "Type": "AWS::Batch::JobQueue",
                "Properties": {
                    "Priority": 1,
                    "ComputeEnvironmentOrder": [
                        {
                            "Order": 1,
                            "ComputeEnvironment": {"Ref": "ComputeEnvironment"},
                        }
                    ],
                },
            },
            "JobDefinition": {
                "Type": "AWS::Batch::JobDefinition",
                "Properties": {
                    "Type": "container",
                    "ContainerProperties": {
                        "Image": {
                            "Fn::Join": [
                                "",
                                [
                                    "137112412989.dkr.ecr.",
                                    {"Ref": "AWS::Region"},
                                    ".amazonaws.com/amazonlinux:latest",
                                ],
                            ]
                        },
                        "ResourceRequirements": [
                            {"Type": "MEMORY", "Value": 2000},
                            {"Type": "VCPU", "Value": 2},
                        ],
                        "Command": ["echo", "Hello world"],
                        "LinuxParameters": {"Devices": [{"HostPath": "test-path"}]},
                    },
                    "RetryStrategy": {"Attempts": 1},
                },
            },
        }
    }
    cf_json = json.dumps(create_environment_template)

    cf_conn = boto3.client("cloudformation", DEFAULT_REGION)
    stack_name = str(uuid4())[0:6]
    stack_id = cf_conn.create_stack(StackName=stack_name, TemplateBody=cf_json)[
        "StackId"
    ]

    stack_resources = cf_conn.list_stack_resources(StackName=stack_id)
    assert len(stack_resources["StackResourceSummaries"]) == 3

    job_def_resource = list(
        filter(
            lambda item: item["ResourceType"] == "AWS::Batch::JobDefinition",
            stack_resources["StackResourceSummaries"],
        )
    )[0]

    assert job_def_resource["ResourceStatus"] == "CREATE_COMPLETE"
    # Spot checks on the ARN
    assert job_def_resource["PhysicalResourceId"].startswith("arn:aws:batch:")
    assert f"{stack_name}-JobDef" in job_def_resource["PhysicalResourceId"]
    assert "job-definition/" in job_def_resource["PhysicalResourceId"]

    # Test the linux parameter device host path
    # This ensures that batch is parsing the parameter dictionaries
    # correctly by recursively converting the first character of all
    # dict keys to lowercase.
    batch_conn = boto3.client("batch", DEFAULT_REGION)
    response = batch_conn.describe_job_definitions(
        jobDefinitions=[job_def_resource["PhysicalResourceId"]]
    )
    job_def_linux_device_host_path = response.get("jobDefinitions")[0][
        "containerProperties"
    ]["linuxParameters"]["devices"][0]["hostPath"]

    assert job_def_linux_device_host_path == "test-path"



"""Exceptions raised by the appmesh service."""

from moto.core.exceptions import JsonRESTError


class MeshError(JsonRESTError):
    code = 400


class MeshNotFoundError(MeshError):
    def __init__(self, mesh_name: str) -> None:
        super().__init__(
            "MeshNotFound",
            f"There are no meshes with the name {mesh_name}.",
        )


class ResourceNotFoundError(MeshError):
    def __init__(self, resource_arn: str) -> None:
        super().__init__(
            "ResourceNotFound",
            f"There are no mesh resources with the arn {resource_arn}.",
        )


class MeshOwnerDoesNotMatchError(MeshError):
    def __init__(self, mesh_name: str, mesh_owner: str) -> None:
        super().__init__(
            "MeshOwnerDoesNotMatch",
            f"The owner of the mesh {mesh_name} does not match the owner name provided: {mesh_owner}.",
        )


class VirtualRouterNameAlreadyTakenError(MeshError):
    def __init__(self, mesh_name: str, virtual_router_name: str) -> None:
        super().__init__(
            "VirtualRouterNameAlreadyTaken",
            f"There is already a virtual router named {virtual_router_name} associated with the mesh {mesh_name}.",
        )


class VirtualRouterNotFoundError(MeshError):
    def __init__(self, mesh_name: str, virtual_router_name: str) -> None:
        super().__init__(
            "VirtualRouterNotFound",
            f"The mesh {mesh_name} does not have a virtual router named {virtual_router_name}.",
        )


class RouteNotFoundError(MeshError):
    def __init__(
        self, mesh_name: str, virtual_router_name: str, route_name: str
    ) -> None:
        super().__init__(
            "RouteNotFound",
            f"There is no route named {route_name} associated with router {virtual_router_name} in mesh {mesh_name}.",
        )


class RouteNameAlreadyTakenError(MeshError):
    def __init__(
        self, mesh_name: str, virtual_router_name: str, route_name: str
    ) -> None:
        super().__init__(
            "RouteNameAlreadyTaken",
            f"There is already a route named {route_name} associated with router {virtual_router_name} in mesh {mesh_name}.",
        )


class MissingRequiredFieldError(MeshError):
    def __init__(self, field_name: str) -> None:
        super().__init__(
            "MissingRequiredField",
            f"{field_name} must be defined.",
        )


class VirtualNodeNotFoundError(MeshError):
    def __init__(self, mesh_name: str, virtual_node_name: str) -> None:
        super().__init__(
            "VirtualNodeNotFound",
            f"{virtual_node_name} is not a virtual node associated with mesh {mesh_name}",
        )


class VirtualNodeNameAlreadyTakenError(MeshError):
    def __init__(self, mesh_name: str, virtual_node_name: str) -> None:
        super().__init__(
            "VirtualNodeNameAlreadyTaken",
            f"There is already a virtual node named {virtual_node_name} associated with mesh {mesh_name}",
        )



import os
from unittest import SkipTest, mock

import boto3

from moto import mock_aws, settings

test_ami = "/aws/service/ami-amazon-linux-latest/al2023-ami-kernel-default-x86_64"


# The default AMIs are not loaded for our test case, to speed things up
# But we do need it for this specific test
@mock.patch.dict(os.environ, {"MOTO_EC2_LOAD_DEFAULT_AMIS": "true"})
@mock_aws
def test_ssm_get_latest_ami_by_path():
    if settings.TEST_SERVER_MODE:
        raise SkipTest("Can't set environment variables in ServerMode")
    ssm = boto3.client("ssm", region_name="us-east-1")
    ami = ssm.get_parameter(Name=test_ami)["Parameter"]["Value"]

    ec2 = boto3.client("ec2", region_name="us-east-1")
    assert len(ec2.describe_images(ImageIds=[ami])["Images"]) == 1



"""Unit tests for directconnect-supported APIs."""

import time

import boto3
import pytest

from moto import mock_aws


@pytest.fixture(name="client")
def fixture_dx_client():
    with mock_aws():
        yield boto3.client("directconnect", region_name="us-east-1")


def test_create_connection(client):
    connection = client.create_connection(
        location="EqDC2", bandwidth="10Gbps", connectionName="TestConnection"
    )
    assert connection["connectionId"].startswith("dx-moto")
    assert connection["connectionState"] == "available"


def test_describe_connections(client):
    client.create_connection(
        location="EqDC2",
        bandwidth="10Gbps",
        connectionName="TestConnection1",
        requestMACSec=False,
    )
    time.sleep(0.1)
    client.create_connection(
        location="EqDC2",
        bandwidth="10Gbps",
        connectionName="TestConnection2",
        requestMACSec=True,
    )
    time.sleep(0.1)
    resp = client.describe_connections()
    connections = resp["connections"]
    assert len(connections) == 2
    assert not connections[0]["macSecCapable"]
    assert connections[1]["macSecCapable"]
    assert len(connections[0]["macSecKeys"]) == 0
    assert len(connections[1]["macSecKeys"]) == 1
    assert connections[0]["encryptionMode"] == "no_encrypt"
    assert connections[1]["encryptionMode"] == "must_encrypt"
    resp = client.describe_connections(connectionId=connections[0]["connectionId"])
    assert len(resp["connections"]) == 1


def test_delete_connection(client):
    connection = client.create_connection(
        location="EqDC2", bandwidth="10Gbps", connectionName="TestConnection"
    )
    connection = client.delete_connection(connectionId=connection["connectionId"])
    assert connection["connectionState"] == "deleted"


def test_update_connection(client):
    client.create_connection(
        location="EqDC2",
        bandwidth="10Gbps",
        connectionName="TestConnection1",
    )
    time.sleep(0.1)
    client.create_connection(
        location="EqDC2",
        bandwidth="10Gbps",
        connectionName="TestConnection2",
        requestMACSec=True,
    )
    time.sleep(0.1)
    resp = client.describe_connections()
    connection1_id = resp["connections"][0]["connectionId"]
    connection2_id = resp["connections"][1]["connectionId"]
    connection = client.update_connection(
        connectionId=connection1_id,
        connectionName="NewConnectionName",
    )
    assert connection["connectionName"] == "NewConnectionName"
    connection = client.update_connection(
        connectionId=connection2_id,
        encryptionMode="should_encrypt",
    )
    assert connection["encryptionMode"] == "should_encrypt"



"""sagemakermetrics base URL and path."""

from .responses import SageMakerMetricsResponse

url_bases = [
    r"https?://metrics.sagemaker\.(.+)\.amazonaws\.com",
]

url_paths = {
    "{0}/$": SageMakerMetricsResponse.dispatch,
    "{0}/BatchPutMetrics$": SageMakerMetricsResponse.dispatch,
}



from moto.core.exceptions import RESTError


class InvalidFormat(RESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__(InvalidFormat.__name__, message)


class InvalidParameterValue(RESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__(InvalidParameterValue.__name__, message)


class InvalidParameterCombination(RESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__(InvalidParameterCombination.__name__, message)


class ResourceNotFound(RESTError):
    code = 404

    def __init__(self) -> None:
        super().__init__(ResourceNotFound.__name__, "Unknown")


class ResourceNotFoundException(RESTError):
    code = 404

    def __init__(self) -> None:
        super().__init__(ResourceNotFoundException.__name__, "Unknown")


class ValidationError(RESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__(ValidationError.__name__, message)



import datetime

import pytest
from freezegun import freeze_time

from moto.core.utils import (
    camelcase_to_pascal,
    camelcase_to_underscores,
    pascal_to_camelcase,
    rfc_1123_datetime,
    str_to_rfc_1123_datetime,
    underscores_to_camelcase,
    unix_time,
)


@pytest.mark.parametrize(
    "_input,expected",
    [
        ("theNewAttribute", "the_new_attribute"),
        ("attri bute With Space", "attribute_with_space"),
        ("FirstLetterCapital", "first_letter_capital"),
        ("ListMFADevices", "list_mfa_devices"),
    ],
)
def test_camelcase_to_underscores(_input: str, expected: str) -> None:
    assert camelcase_to_underscores(_input) == expected


@pytest.mark.parametrize(
    "_input,expected",
    [("the_new_attribute", "theNewAttribute"), ("attribute", "attribute")],
)
def test_underscores_to_camelcase(_input: str, expected: str) -> None:
    assert underscores_to_camelcase(_input) == expected


@pytest.mark.parametrize(
    "_input,expected",
    [("TheNewAttribute", "theNewAttribute"), ("Attribute", "attribute")],
)
def test_pascal_to_camelcase(_input: str, expected: str) -> None:
    assert pascal_to_camelcase(_input) == expected


@pytest.mark.parametrize(
    "_input,expected",
    [("theNewAttribute", "TheNewAttribute"), ("attribute", "Attribute")],
)
def test_camelcase_to_pascal(_input: str, expected: str) -> None:
    assert camelcase_to_pascal(_input) == expected


@freeze_time("2015-01-01 12:00:00")
def test_unix_time() -> None:  # type: ignore[misc]
    assert unix_time() == 1420113600.0


def test_rfc1123_dates() -> None:
    with freeze_time("2012-03-04 05:00:05"):
        x = rfc_1123_datetime(datetime.datetime.now())
        assert x == "Sun, 04 Mar 2012 05:00:05 GMT"
        assert str_to_rfc_1123_datetime(x) == datetime.datetime.now()
